# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
# import PySide6  # noqa: F401
# from PySide6.QtWidgets import QLabel, QPushButton, QVBoxLayout, QWidget
from PySide6.QtGui import QTransform

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.obstacle import Detector, DetectorConfig, DetectorResult


SENSOR_IDS = [2, 3]


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client.open(**a121.get_client_args(args))

    # Creating a list of subsweep configurations.
    subsweep_configurations = [
        a121.SubsweepConfig(
            start_point=24,  # ~6 cm
            num_points=96,
            step_length=1,
            profile=a121.Profile.PROFILE_1,
            hwaas=4,
        ),
        a121.SubsweepConfig(
            start_point=96,  # ~24 cm
            num_points=48,
            step_length=4,
            profile=a121.Profile.PROFILE_3,
            hwaas=16,
        ),
        a121.SubsweepConfig(
            start_point=300,  # ~75 cm
            num_points=48,
            step_length=4,
            profile=a121.Profile.PROFILE_3,
            hwaas=2,
        ),
    ]

    # Configure the detector using multiple subsweeps
    detector_config = DetectorConfig(
        enable_bilateration=(len(SENSOR_IDS) == 2),
        bilateration_sensor_spacing_m=0.1,
        num_mean_threshold=1.5,
        num_std_threshold=4,
        subsweep_configurations=subsweep_configurations,
    )

    detector = Detector(client=client, sensor_ids=SENSOR_IDS, detector_config=detector_config)

    detector.calibrate_detector()
    detector.start()

    pg_updater = PGUpdater(
        num_sensors=len(SENSOR_IDS),
        num_subsweeps=len(subsweep_configurations),
        enable_bilateration=detector_config.enable_bilateration,
    )

    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        # v_current = get_speed_from_some_Robot_API(), e.g ROS?
        # detector.update_robot_speed(v_current)

        detector_result = detector.get_next()

        try:
            pg_process.put_data(detector_result)
        except et.PGProccessDiedException:
            break

    detector.stop()

    print("Disconnecting...")
    client.close()


PLOT_HISTORY_FRAMES = 50
PLOT_THRESHOLDS = True


class PGUpdater:
    def __init__(self, num_sensors, num_subsweeps, enable_bilateration):
        self.num_sensors = num_sensors
        self.num_subsweeps = num_subsweeps
        self.enable_bilateration = enable_bilateration
        self.obst_vel_ys = num_sensors * [np.nan * np.ones(PLOT_HISTORY_FRAMES)]
        self.obst_dist_ys = num_sensors * [np.nan * np.ones(PLOT_HISTORY_FRAMES)]
        self.obst_bil_ys = np.nan * np.ones(PLOT_HISTORY_FRAMES)
        self.hist_x = np.linspace(-100, 0, PLOT_HISTORY_FRAMES)

    def setup(self, win: pg.GraphicsLayout):
        self.fftmap_plots: list[pg.PlotItem] = []
        self.fftmap_images: list[pg.ImageItem] = []
        self.range_hist_curves: list[pg.PlotDataItem] = []
        self.angle_hist_curves: list[pg.PlotDataItem] = []

        if PLOT_THRESHOLDS:
            self.bin0_curves: list[pg.PlotDataItem] = []
            self.bin0_threshold_curves: list[pg.PlotDataItem] = []
            self.other_bins_curves: list[pg.PlotDataItem] = []
            self.other_bins_threshold_curves: list[pg.PlotDataItem] = []

        row_offset = 2 if PLOT_THRESHOLDS else 0

        for i_s in range(self.num_sensors):
            for i_ss in range(self.num_subsweeps):
                col = i_s * self.num_subsweeps + i_ss
                p = win.addPlot(
                    row=0, col=col, title=f"FFT Map, sensor {SENSOR_IDS[i_s]}, subsweep {i_ss}"
                )
                im = pg.ImageItem(autoDownsample=True)
                im.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
                self.fftmap_images.append(im)

                p.setLabel("bottom", "Distance (cm)")
                p.setLabel("left", "Velocity (cm/s)")
                p.addItem(im)

                self.fftmap_plots.append(p)

            if PLOT_THRESHOLDS:
                self.bin0 = pg.PlotItem(title="Zeroth velocity/angle bin")
                self.bin0.showGrid(x=True, y=True)
                self.bin0.setLabel("bottom", "Range (cm)")
                self.bin0.setLabel("left", "Amplitude")
                self.bin0.addLegend()

                pen = et.utils.pg_pen_cycler(0)
                brush = et.utils.pg_brush_cycler(0)
                symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
                feat_kw = dict(pen=pen, **symbol_kw)
                self.bin0_curves += [self.bin0.plot(**feat_kw) for _ in range(self.num_subsweeps)]

                pen = et.utils.pg_pen_cycler(1)
                brush = et.utils.pg_brush_cycler(1)
                symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
                feat_kw = dict(pen=pen, **symbol_kw)
                self.bin0_threshold_curves += [
                    self.bin0.plot(**feat_kw) for _ in range(self.num_subsweeps)
                ]

                bin0_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
                bin0_plot_legend.setParentItem(self.bin0)
                bin0_plot_legend.addItem(self.bin0_curves[0], "Sweep")
                bin0_plot_legend.addItem(self.bin0_threshold_curves[0], "Threshold")

                sublayout = win.addLayout(
                    row=1,
                    col=i_s * self.num_subsweeps,
                    colspan=self.num_subsweeps,
                )
                sublayout.layout.setColumnStretchFactor(0, self.num_subsweeps)
                sublayout.addItem(self.bin0, row=0, col=0)

                self.other_bins = pg.PlotItem(title="Other velocity/angle bins")
                self.other_bins.showGrid(x=True, y=True)
                self.other_bins.setLabel("bottom", "Range (cm)")
                self.other_bins.setLabel("left", "Amplitude")
                self.other_bins.addLegend()

                pen = et.utils.pg_pen_cycler(0)
                brush = et.utils.pg_brush_cycler(0)
                symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
                feat_kw = dict(pen=pen, **symbol_kw)
                self.other_bins_curves += [
                    self.other_bins.plot(**feat_kw) for _ in range(self.num_subsweeps)
                ]

                pen = et.utils.pg_pen_cycler(1)
                brush = et.utils.pg_brush_cycler(1)
                symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
                feat_kw = dict(pen=pen, **symbol_kw)
                self.other_bins_threshold_curves += [
                    self.other_bins.plot(**feat_kw) for _ in range(self.num_subsweeps)
                ]

                other_bins_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
                other_bins_plot_legend.setParentItem(self.other_bins)
                other_bins_plot_legend.addItem(self.other_bins_curves[0], "Max Sweep")
                other_bins_plot_legend.addItem(self.other_bins_threshold_curves[0], "Threshold")

                sublayout = win.addLayout(
                    row=2,
                    col=i_s * self.num_subsweeps,
                    colspan=self.num_subsweeps,
                )
                sublayout.layout.setColumnStretchFactor(0, self.num_subsweeps)
                sublayout.addItem(self.other_bins, row=0, col=0)

            self.angle_hist = pg.PlotItem(title="Angle/velocity history")
            self.angle_hist.showGrid(x=True, y=True)
            self.angle_hist.setLabel("bottom", "Time (frames)")
            self.angle_hist.setLabel("left", "velocity (cm/s)")
            self.angle_hist.setXRange(-100, 0)
            self.angle_hist.addLegend()
            self.angle_hist_curves.append(self.angle_hist.plot(symbolSize=5, symbol="o"))

            sublayout = win.addLayout(
                row=1 + row_offset,
                col=i_s * self.num_subsweeps,
                colspan=self.num_subsweeps,
            )

            sublayout.layout.setColumnStretchFactor(0, self.num_subsweeps)
            sublayout.addItem(self.angle_hist, row=0, col=0)

            self.range_hist = pg.PlotItem(title="Range history")
            self.range_hist.showGrid(x=True, y=True)
            self.range_hist.setLabel("bottom", "Time (frames)")
            self.range_hist.setLabel("left", "Range (cm)")
            self.range_hist.setXRange(-100, 0)
            self.range_hist.addLegend()
            self.range_hist_curves.append(self.range_hist.plot(symbolSize=5, symbol="o"))

            sublayout = win.addLayout(
                row=2 + row_offset,
                col=i_s * self.num_subsweeps,
                colspan=self.num_subsweeps,
            )

            sublayout.layout.setColumnStretchFactor(0, self.num_subsweeps)
            sublayout.addItem(self.range_hist, row=0, col=0)

        if self.enable_bilateration:
            self.bil_hist_plot = pg.PlotItem(title="Bilateration history")

            self.bil_hist_plot.showGrid(x=True, y=True)
            self.bil_hist_plot.setLabel("bottom", "Time (frames)")
            self.bil_hist_plot.setLabel("left", "Bilateration angle (deg)")
            self.bil_hist_plot.setXRange(-100, 0)
            self.bil_hist_plot.setYRange(-90, 90)
            self.bil_hist_plot.addLegend()

            self.bil_hist_curve = self.bil_hist_plot.plot(pen=et.utils.pg_pen_cycler(1))

            sublayout = win.addLayout(row=3 + row_offset, col=0, colspan=2 * self.num_subsweeps)
            sublayout.layout.setColumnStretchFactor(0, 2 * self.num_subsweeps)
            sublayout.addItem(self.bil_hist_plot, row=0, col=0)

    def update(self, detector_result: DetectorResult):
        for i_s in range(self.num_sensors):
            pr = detector_result.processor_results[SENSOR_IDS[i_s]]

            for i_ss in range(self.num_subsweeps):
                curve_idx = self.num_subsweeps * i_s + i_ss

                fftmap = pr.subsweeps_extra_results[i_ss].fft_map
                fftmap_threshold = pr.subsweeps_extra_results[i_ss].fft_map_threshold

                # fftmap = 10*np.log10(fftmap)  # to dB

                spf = fftmap.shape[0]
                r = 100 * pr.subsweeps_extra_results[i_ss].r

                transform = QTransform()
                transform.translate(
                    r[0], -100 * pr.extra_result.dv * spf / 2 - 0.5 * 100 * pr.extra_result.dv
                )
                transform.scale(r[1] - r[0], 100 * pr.extra_result.dv)

                self.fftmap_images[curve_idx].setTransform(transform)

                self.fftmap_images[curve_idx].updateImage(
                    np.fft.fftshift(fftmap, 0).T,
                    levels=(0, 1.05 * np.max(fftmap)),
                )

                if PLOT_THRESHOLDS:
                    bin0 = fftmap[0, :]
                    threshold_bin0 = fftmap_threshold[0, :]

                    max_other_bins = np.max(fftmap[1:, :], axis=0)
                    threshold_other_bins = fftmap_threshold[1, :]

                    self.bin0_curves[curve_idx].setData(r, bin0)
                    self.bin0_threshold_curves[curve_idx].setData(r, threshold_bin0)
                    self.other_bins_curves[curve_idx].setData(r, max_other_bins)
                    self.other_bins_threshold_curves[curve_idx].setData(r, threshold_other_bins)

            v = pr.targets[0].velocity if pr.targets else np.nan

            self.obst_vel_ys[i_s] = np.roll(self.obst_vel_ys[i_s], -1)
            self.obst_vel_ys[i_s][-1] = 100 * v  # m/s -> cm/s

            if np.isnan(self.obst_vel_ys[i_s]).all():
                self.angle_hist_curves[i_s].setVisible(False)
            else:
                self.angle_hist_curves[i_s].setVisible(True)
                self.angle_hist_curves[i_s].setData(
                    self.hist_x, self.obst_vel_ys[i_s], connect="finite"
                )

            # TODO: should be known earlier
            # r_min = min([er.r[0] for er in pr.subsweeps_extra_results])
            # r_max = min([er.r[-1] for er in pr.subsweeps_extra_results])
            # self.range_hist.setYRange(100*r_min, 100*r_max)

            r_targets = pr.targets[0].distance if pr.targets else np.nan

            self.obst_dist_ys[i_s] = np.roll(self.obst_dist_ys[i_s], -1)
            self.obst_dist_ys[i_s][-1] = 100 * r_targets  # m -> cm

            # print(f'{i_s}: r = {r_targets}, v = {v}')

            if np.isnan(self.obst_dist_ys[i_s]).all():
                self.range_hist_curves[i_s].setVisible(False)
            else:
                self.range_hist_curves[i_s].setVisible(True)
                self.range_hist_curves[i_s].setData(
                    self.hist_x, self.obst_dist_ys[i_s], connect="finite"
                )

        if self.enable_bilateration:
            beta = (
                detector_result.bilateration_result.beta_degs[0]
                if detector_result.bilateration_result.beta_degs
                else np.nan
            )

            self.obst_bil_ys = np.roll(self.obst_bil_ys, -1)
            self.obst_bil_ys[-1] = beta

            if np.isnan(self.obst_bil_ys).all():
                self.bil_hist_curve.setVisible(False)
            else:
                self.bil_hist_curve.setVisible(True)
                self.bil_hist_curve.setData(self.hist_x, self.obst_bil_ys, connect="finite")


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from ._bilaterator import (
    Bilaterator,
    BilateratorResult,
)
from ._detector import DetailedStatus, Detector, DetectorConfig, DetectorContext, DetectorResult
from ._processors import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
    Target,
)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from acconeer.exptool import a121

from ._detector import DetectorConfig


def get_short_range_config() -> DetectorConfig:
    return DetectorConfig(
        start_m=0.06,
        end_m=1.0,
        frame_rate=10,
        sweeps_per_frame=16,
        hwaas=16,
        inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
        intra_enable=True,
        intra_detection_threshold=1.4,
        intra_frame_time_const=0.15,
        intra_output_time_const=0.3,
        inter_enable=True,
        inter_detection_threshold=1,
        inter_frame_slow_cutoff=0.2,
        inter_frame_fast_cutoff=5,
        inter_frame_deviation_time_const=0.5,
        inter_output_time_const=2,
        inter_phase_boost=False,
        inter_frame_presence_timeout=3,
    )


def get_medium_range_config() -> DetectorConfig:
    return DetectorConfig(
        start_m=0.3,
        end_m=2.5,
        frame_rate=12,
        sweeps_per_frame=16,
        hwaas=32,
        inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
        intra_enable=True,
        intra_detection_threshold=1.3,
        intra_frame_time_const=0.15,
        intra_output_time_const=0.3,
        inter_enable=True,
        inter_detection_threshold=1,
        inter_frame_slow_cutoff=0.2,
        inter_frame_fast_cutoff=6,
        inter_frame_deviation_time_const=0.5,
        inter_output_time_const=2,
        inter_phase_boost=False,
        inter_frame_presence_timeout=3,
    )


def get_long_range_config() -> DetectorConfig:
    return DetectorConfig(
        start_m=5,
        end_m=7.5,
        frame_rate=12,
        sweeps_per_frame=32,
        hwaas=128,
        inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
        intra_enable=True,
        intra_detection_threshold=1.2,
        intra_frame_time_const=0.15,
        intra_output_time_const=0.3,
        inter_enable=True,
        inter_detection_threshold=0.8,
        inter_frame_slow_cutoff=0.2,
        inter_frame_fast_cutoff=6,
        inter_frame_deviation_time_const=0.5,
        inter_output_time_const=2,
        inter_phase_boost=False,
        inter_frame_presence_timeout=3,
    )


def get_low_power_config() -> DetectorConfig:
    return DetectorConfig(
        start_m=0.725,
        end_m=1.505,
        frame_rate=1,
        sweeps_per_frame=16,
        hwaas=8,
        inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
        intra_enable=True,
        intra_detection_threshold=1.5,
        intra_frame_time_const=0.15,
        intra_output_time_const=0.3,
        inter_enable=True,
        inter_detection_threshold=1,
        inter_frame_slow_cutoff=0.2,
        inter_frame_fast_cutoff=5,
        inter_frame_deviation_time_const=0.5,
        inter_output_time_const=2,
        inter_phase_boost=False,
        inter_frame_presence_timeout=3,
    )



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import copy
import enum
import warnings
from typing import Any, Dict, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121, opser
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    AlgoBase,
    AlgoConfigBase,
    Controller,
)
from acconeer.exptool.a121.algo._utils import get_distances_m
from acconeer.exptool.a121.algo.parking._processors import (
    ObstructionProcessor,
    ObstructionProcessorConfig,
    Processor,
    ProcessorConfig,
)


opser.register_json_presentable(a121.SensorConfig)


class DetailedStatus(enum.Enum):
    OK = enum.auto()
    CALIBRATION_MISSING = enum.auto()
    CONFIG_MISMATCH = enum.auto()
    OBSTRUCTION_CALIBRATION_MISSING = enum.auto()


@attrs.frozen(kw_only=True)
class RefAppStatus:
    ref_app_state: DetailedStatus
    ready_to_start: bool


@attrs.mutable(kw_only=True)
class RefAppContext(AlgoBase):
    noise_level: float = attrs.field(default=1.0)
    obstruction_center: npt.NDArray[np.float_] = attrs.field(default=np.array([0.0, 0.0]))
    calibration_temperature: float = attrs.field(default=0.0)
    calibration_done: bool = attrs.field(default=False)
    calibration_sensor_config: a121.SensorConfig = attrs.field(default=a121.SensorConfig())
    obstruction_calibration_done: bool = attrs.field(default=False)
    obstruction_noise_level: float = attrs.field(default=0.0)
    obstruction_sensor_config: a121.SensorConfig = attrs.field(default=a121.SensorConfig())


@attributes_doc
@attrs.mutable(kw_only=True)
class RefAppConfig(AlgoConfigBase):
    range_start_m: float = attrs.field(default=0.3)
    """Start of measurement range (m)."""

    range_end_m: float = attrs.field(default=1.5)
    """End of measurement range (m)."""

    hwaas: int = attrs.field(default=32)
    """HWAAS."""

    profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_2, converter=a121.Profile)
    """Profile."""

    update_rate: float = attrs.field(default=10.0)
    """Frame rate."""

    queue_length_n: int = attrs.field(default=5)
    """Number of consecutive samples stored."""

    amplitude_threshold: float = attrs.field(default=15.0)
    """Threshold level in times the noise level to trigger."""

    weighted_distance_threshold_m: float = attrs.field(default=0.1)
    """Max distance between two signatures to consider samples similar."""

    signature_similarity_threshold: Optional[float] = attrs.field(default=60.0)
    """How large fraction of signatures have to be within distance threshold to detect car."""

    obstruction_detection: bool = attrs.field(default=False)
    """If obstruction detection should be active. This costs some extra battery power."""

    obstruction_distance_threshold: float = attrs.field(default=0.06)
    """Distance in similarity space of obstruction signature to consider sensor obstructed."""

    obstruction_start_m: float = attrs.field(default=0.03)
    """Start of obstruction detection range."""

    obstruction_end_m: float = attrs.field(default=0.05)
    """End of obstruction detection range."""

    frame_noise_estimation: bool = attrs.field(default=False)
    """Active noise estimation every frame. Replaces calibration at cost of frame size."""

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> RefAppConfig:
        return cls(**d)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []
        fwhm = ENVELOPE_FWHM_M[self.profile]

        if self.range_start_m < 2 * fwhm:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "range_start_m",
                    "Start point inside direct leakage",
                )
            )

        if self.range_end_m < self.range_start_m:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "range_end_m",
                    "End point must be greater than start point",
                )
            )

        if self.range_end_m > 5.0:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "range_end_m",
                    "End point must be within 5m, see documentation for details.",
                )
            )
        return validation_results


@attrs.frozen(kw_only=True)
class RefAppExtraResult:
    signature_history: npt.NDArray[np.float_] = attrs.field()
    """Array containing queue_length_n last signatures."""

    parking_data: npt.NDArray[np.float_] = attrs.field()
    """The scaled amplitude array used to calculate the last signature."""

    closest_object_dist: float = attrs.field(default=0.0)
    """Location of closest signature in cluster. Only availible if car is detected."""

    obstruction_data: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    """Array with obstruction amplitudes (usually direct leakage)."""

    obstruction_signature: npt.NDArray[np.float_] = attrs.field(default=(0.0, 0.0))
    """Signature of amplitudes used for the obstruction detection."""

    obstruction_center: npt.NDArray[np.float_] = attrs.field(default=(0.0, 0.0))
    """Signature of amplitudes used in calibration."""

    obstruction_distance: float = attrs.field(default=0.0)
    """Obstruction distance threshold, same as in configuration."""


@attrs.frozen(kw_only=True)
class RefAppResult:
    car_detected: bool = attrs.field(default=False)
    """Boolean indicating if a car is present."""

    obstruction_detected: bool = attrs.field(default=False)
    """Boolean indicating whether something is obstructing the sensor."""

    extra_result: RefAppExtraResult = attrs.field()
    """Extra information for plotting."""


class RefApp(Controller[RefAppConfig, RefAppResult]):
    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        ref_app_config: RefAppConfig,
        context: RefAppContext = RefAppContext(),
    ) -> None:
        super().__init__(client=client, config=ref_app_config)

        self.all_sensor_configs = get_sensor_configs(ref_app_config)
        self.sensor_config = self.all_sensor_configs["full_config"]
        self.sensor_id = sensor_id
        self.n_samples_to_calibrate = 3  # number of samples used for calibration.

        groups = [{sensor_id: self.sensor_config}]
        if ref_app_config.obstruction_detection:
            self.obstruction_config = self.all_sensor_configs["obstruction_config"]
            groups.append({sensor_id: self.obstruction_config})
        self.session_config = a121.SessionConfig(groups, update_rate=ref_app_config.update_rate)

        if ref_app_config.signature_similarity_threshold is not None:
            # input is a percentage, divide to get useful value
            signature_similarity_threshold = ref_app_config.signature_similarity_threshold / 100.0
        else:
            signature_similarity_threshold = 0.6

        self.processor_config = ProcessorConfig(
            queue_length=ref_app_config.queue_length_n,
            amplitude_threshold=ref_app_config.amplitude_threshold,
            frame_noise_estimation=ref_app_config.frame_noise_estimation,
            weighted_distance_threshold_m=ref_app_config.weighted_distance_threshold_m,
            signature_similarity_threshold=signature_similarity_threshold,
        )
        self.ref_app_config = ref_app_config
        self.sensor_id = sensor_id

        self.context = context

        self.noise_level = self.context.noise_level  # noise level estimated at calibration
        self.calibration_temperature = self.context.calibration_temperature

        if ref_app_config.obstruction_detection:
            self.obstruction_center = self.context.obstruction_center
            self.obstruction_noise_level = self.context.obstruction_noise_level

        self.calibration_done = False
        self.started = False

    def start(
        self,
        recorder: Optional[a121.Recorder] = None,
        algo_group: Optional[h5py.Group] = None,
    ) -> None:
        if self.started:
            raise RuntimeError("Already started")

        self.all_metadata = self.client.setup_session(self.session_config)

        if self.ref_app_config.obstruction_detection:
            assert isinstance(self.all_metadata, list)
            self.metadata = self.all_metadata[0][self.sensor_id]
            self.obs_metadata = self.all_metadata[1][self.sensor_id]
        else:
            assert isinstance(self.all_metadata, a121.Metadata)
            self.metadata = self.all_metadata

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                if algo_group is None:
                    algo_group = recorder.require_algo_group("parking")
                _record_algo_data(
                    algo_group,
                    self.sensor_id,
                    self.config,
                    self.context,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

            self.client.attach_recorder(recorder)

        base_sensor_config = self.all_sensor_configs["full_config"]

        self.base_processor = Processor(
            sensor_config=base_sensor_config,
            processor_config=self.processor_config,
            metadata=self.metadata,
            noise_estimate=self.noise_level,
            noise_estimate_temperature=self.calibration_temperature,
        )

        if self.ref_app_config.obstruction_detection:
            obstruction_sensor_config = self.all_sensor_configs["obstruction_config"]
            obstruction_distance_threshold = self.ref_app_config.obstruction_distance_threshold
            obstruction_processor_config = ObstructionProcessorConfig(
                distance_threshold=obstruction_distance_threshold
            )
            self.obs_distances = get_distances_m(obstruction_sensor_config, self.obs_metadata)

            self.obstruction_processor = ObstructionProcessor(
                sensor_config=obstruction_sensor_config,
                processor_config=obstruction_processor_config,
                metadata=self.metadata,
                update_rate=self.ref_app_config.update_rate,
                calibration_center=self.obstruction_center,
                calibration_noise_mean=self.obstruction_noise_level,
                calibration_temperature=self.calibration_temperature,
            )

        self.client.start_session()

        self.started = True

    def calibrate_ref_app(self) -> None:
        obstruction_detection = self.ref_app_config.obstruction_detection

        temp_config = copy.deepcopy(self.ref_app_config)
        temp_config.frame_noise_estimation = False
        sensor_configs = get_sensor_configs(temp_config)
        sensor_config = sensor_configs["full_config"]

        sensor_config.enable_tx = False

        if obstruction_detection:
            # We need two calibration values, one to determine the signature and one to determine the noise level.
            obstruction_config = sensor_configs["obstruction_config"]
            obstruction_config_tx_off = copy.deepcopy(obstruction_config)
            obstruction_config_tx_off.enable_tx = False
            assert isinstance(obstruction_config, a121.SensorConfig)
            groups = [
                {self.sensor_id: sensor_config},
                {self.sensor_id: obstruction_config},
                {self.sensor_id: obstruction_config_tx_off},
            ]
            session_config = a121.SessionConfig(groups, update_rate=self.n_samples_to_calibrate)
        else:
            session_config = a121.SessionConfig(
                {self.sensor_id: sensor_config}, update_rate=self.n_samples_to_calibrate
            )

        metadata = self.client.setup_session(session_config)

        if obstruction_detection:
            assert isinstance(metadata, list)
            obs_metadata = metadata[1][self.sensor_id]
            obs_distances = get_distances_m(obstruction_config, obs_metadata)

        noise_levels = []
        obs_noise_levels = []
        signatures = []
        temperatures = []

        self.client.start_session()
        for i in range(self.n_samples_to_calibrate):
            main_result = self.client.get_next()
            if obstruction_detection:
                assert isinstance(main_result, list)
                result = main_result[0][self.sensor_id]
                obs_result = main_result[1][self.sensor_id]
                obs_tx_off_result = main_result[2][self.sensor_id]
                noise_level = ObstructionProcessor.get_noise_level(obs_tx_off_result.frame)
                obs_noise_levels.append(noise_level)
                signature, _ = ObstructionProcessor.get_signature(
                    obs_result.frame, noise_level, obs_distances
                )
                signatures.append(signature)
            else:
                assert isinstance(main_result, a121.Result)
                result = main_result

            temperature = result.temperature
            temperatures.append(temperature)

            noise_frame = result.frame
            noise_level = Processor.process_noise_frame(noise_frame)
            noise_levels.append(noise_level)

        if obstruction_detection:
            self.obstruction_center = np.mean(signatures, axis=0)
            self.context.obstruction_noise_level = float(np.mean(obs_noise_levels))
            self.context.obstruction_center = self.obstruction_center
            self.context.obstruction_calibration_done = True
            self.context.obstruction_sensor_config = obstruction_config
        else:
            self.context.obstruction_calibration_done = False

        self.client.stop_session()

        self.calibration_temperature = float(np.median(temperatures))
        self.noise_level = float(np.mean(noise_levels))

        # Update the context
        sensor_config.enable_tx = True
        self.context.noise_level = self.noise_level
        self.context.calibration_temperature = self.calibration_temperature
        self.context.calibration_sensor_config = sensor_config
        self.context.calibration_done = True

    def get_next(self) -> RefAppResult:
        if not self.started:
            raise RuntimeError("Not started")

        result = self.client.get_next()

        if self.ref_app_config.obstruction_detection:
            assert isinstance(result, list)
            base_res = result[0][self.sensor_id]
            obs_res = result[1][self.sensor_id]
        else:
            assert isinstance(result, a121.Result)
            base_res = result

        processor_result = self.base_processor.process(base_res)

        if self.ref_app_config.obstruction_detection:
            obstruction_result = self.obstruction_processor.process(obs_res)

            obstruction_found = obstruction_result.obstruction_found

            extra_result = RefAppExtraResult(
                signature_history=processor_result.extra_result.signature_history,
                parking_data=processor_result.extra_result.parking_data,
                closest_object_dist=processor_result.extra_result.closest_observation,
                obstruction_data=obstruction_result.extra_result.obstruction_data,
                obstruction_signature=obstruction_result.extra_result.obstruction_signature,
                obstruction_center=obstruction_result.extra_result.obstruction_center,
                obstruction_distance=obstruction_result.extra_result.obstruction_distance,
            )
        else:
            obstruction_found = False
            extra_result = RefAppExtraResult(
                signature_history=processor_result.extra_result.signature_history,
                parking_data=processor_result.extra_result.parking_data,
                closest_object_dist=processor_result.extra_result.closest_observation,
            )

        ref_app_result = RefAppResult(
            car_detected=processor_result.car_detected,
            obstruction_detected=obstruction_found,
            extra_result=extra_result,
        )

        return ref_app_result

    def stop(self) -> Any:
        if not self.started:
            raise RuntimeError("Already stopped")

        self.client.stop_session()
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        self.started = False

        return recorder_result

    @classmethod
    def get_ref_app_status(
        cls, ref_app_config: RefAppConfig, context: RefAppContext
    ) -> RefAppStatus:
        using_calibration = not ref_app_config.frame_noise_estimation
        calibration_done = context.calibration_done

        sensor_settings = get_sensor_configs(ref_app_config)
        sensor_conf = sensor_settings["full_config"]

        config_match = context.calibration_sensor_config == sensor_conf

        if ref_app_config.obstruction_detection:
            obstruction_conf = sensor_settings["obstruction_config"]
            obstruction_config_match = obstruction_conf == context.obstruction_sensor_config
            obstruction_ready = context.obstruction_calibration_done and obstruction_config_match
        else:
            obstruction_ready = True

        if using_calibration:
            if not calibration_done:
                ret = RefAppStatus(
                    ready_to_start=False,
                    ref_app_state=DetailedStatus.CALIBRATION_MISSING,
                )
            elif not config_match:
                ret = RefAppStatus(
                    ready_to_start=False,
                    ref_app_state=DetailedStatus.CONFIG_MISMATCH,
                )
            elif not obstruction_ready:
                ret = RefAppStatus(
                    ready_to_start=False,
                    ref_app_state=DetailedStatus.OBSTRUCTION_CALIBRATION_MISSING,
                )
            else:
                ret = RefAppStatus(ready_to_start=True, ref_app_state=DetailedStatus.OK)
        elif not obstruction_ready:
            ret = RefAppStatus(
                ready_to_start=False,
                ref_app_state=DetailedStatus.OBSTRUCTION_CALIBRATION_MISSING,
            )
        else:
            ret = RefAppStatus(ready_to_start=True, ref_app_state=DetailedStatus.OK)

        return ret


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    ref_app_config: RefAppConfig,
    context: RefAppContext,
) -> None:
    algo_group.create_dataset(
        "ref_app_sensor_id",
        data=sensor_id,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "ref_app_config", ref_app_config.to_json())

    context_group = algo_group.create_group("context")
    opser.serialize(context, context_group)


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, RefAppConfig, RefAppContext]:
    sensor_id = int(algo_group["ref_app_sensor_id"][()])

    config = RefAppConfig.from_json(algo_group["ref_app_config"][()])

    context_group = algo_group["context"]
    context = opser.deserialize(context_group, RefAppContext)

    return sensor_id, config, context


def get_obstruction_sensor_config(ref_app_config: RefAppConfig) -> a121.SensorConfig:
    step_length = 2  # Typically not very important, we just want a few points close to the sensor.

    start_point = int(round(ref_app_config.obstruction_start_m / APPROX_BASE_STEP_LENGTH_M))
    num_points = int(
        round(
            np.ceil(
                (ref_app_config.obstruction_end_m - ref_app_config.obstruction_start_m)
                / (step_length * APPROX_BASE_STEP_LENGTH_M)
            )
        )
    )
    conf_obstruction = a121.SensorConfig(
        start_point=start_point,
        num_points=num_points,
        step_length=step_length,
        profile=a121.Profile.PROFILE_1,
        hwaas=16,
        sweeps_per_frame=16,
    )
    return conf_obstruction


def get_sensor_configs(ref_app_config: RefAppConfig) -> Dict[str, a121.SensorConfig]:
    start_point = int(round(ref_app_config.range_start_m / APPROX_BASE_STEP_LENGTH_M))
    profile = a121.Profile(ref_app_config.profile)

    fwhm = ENVELOPE_FWHM_M[profile]

    step_length_ideal = int(round((fwhm / APPROX_BASE_STEP_LENGTH_M)))
    if step_length_ideal >= 24:
        step_length = int(np.floor(step_length_ideal / 24) * 24)
    else:
        step_length = step_length_ideal
        while step_length > 2 and 24 % step_length != 0:
            step_length -= 1
    hwaas = ref_app_config.hwaas
    num_points = int(
        round(
            np.ceil(
                (ref_app_config.range_end_m - ref_app_config.range_start_m)
                / (step_length * APPROX_BASE_STEP_LENGTH_M)
            )
        )
    )

    subsweeps = []
    conf_base = a121.SubsweepConfig(
        start_point=start_point,
        num_points=num_points,
        step_length=step_length,
        profile=profile,
        enable_tx=True,
        hwaas=hwaas,
    )

    ret = {}
    ret["base_config"] = a121.SensorConfig(subsweeps=[conf_base])

    subsweeps.append(conf_base)
    if ref_app_config.frame_noise_estimation:
        conf_noise = a121.SubsweepConfig(
            start_point=start_point,
            num_points=num_points,
            step_length=step_length,
            profile=profile,
            enable_tx=False,
            hwaas=hwaas,
        )
        subsweeps.append(conf_noise)

    if ref_app_config.obstruction_detection:
        conf_obstruction = get_obstruction_sensor_config(ref_app_config)
        ret["obstruction_config"] = conf_obstruction

    conf = a121.SensorConfig(subsweeps=subsweeps, sweeps_per_frame=3)
    ret["full_config"] = conf

    return ret



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import enum
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    AlgoConfigBase,
    Controller,
)
from acconeer.exptool.a121.algo.presence._processors import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
)


class DetectionState(enum.Enum):
    """Indicating the current detection state."""

    NO_DETECTION = enum.auto()
    """No detection."""

    DETECTION = enum.auto()
    """Detection."""

    RETENTION = enum.auto()
    """No detection. Retaining detection."""


@attrs.mutable(kw_only=True)
class ExampleAppConfig(AlgoConfigBase):
    # Setup parameters
    sensor_to_water_distance: float = attrs.field(default=0.12)
    """Distance from the sensor location to the center of the water jet (m)."""

    water_jet_width: float = attrs.field(default=0.02)
    """Water jet width (m).

    Defines the width of a region, centered around the water jet, where the hand motion is not
    analyzed, in order to exclude the water jet motion from the motion metric.
    """

    measurement_range_end: float = attrs.field(default=0.2)
    """Distance between sensor and end point of measurement range (m)."""

    # Filtering parameters
    filter_time_const: float = attrs.field(default=0.3)
    """Time constant of filter applied during processing.

    A lower value yield a more responsive algorithm. A higher value yield a more robust response.
    """

    threshold: float = attrs.field(default=1.5)
    """Threshold applied to the calculate metric.

    If the metric is above the threshold, a hand movement has been detected.
    """

    detection_retention_duration: float = attrs.field(default=1.0)
    """Detection retention duration (s).

    The detection will be kept as True for this duration, after detection is lost.
    """

    # Sensor config parameters
    hwaas: int = attrs.field(default=32)
    """HWAAS."""

    sweeps_per_frame: int = attrs.field(default=64)
    """Sweeps per frame."""

    sweeps_rate: int = attrs.field(default=1000)
    """Sweep rate."""

    frame_rate: int = attrs.field(default=10)
    """Frame rate."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        return validation_results


@attrs.frozen(kw_only=True)
class ExtraResult:
    history: npt.NDArray[np.float_] = attrs.field()
    """Detection history."""

    history_time: npt.NDArray[np.float_] = attrs.field()
    """Time vector for detection history."""

    threshold: float = attrs.field()
    """Threshold."""


@attrs.frozen(kw_only=True)
class ExampleAppResult:
    detection_state: DetectionState = attrs.field()
    """The state of the detection."""

    extra_result: ExtraResult = attrs.field()
    """Extra result used for plotting."""


class ExampleApp(Controller[ExampleAppConfig, ExampleAppResult]):
    """Hand motion detection example app.

    :param client: Client
    :param sensor_id: Sensor id
    :param example_app_config: Hand motion detection configuration
    """

    processors: list[Processor]
    HISTORY_LENGTH_S = 10.0

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        example_app_config: ExampleAppConfig,
    ) -> None:
        super().__init__(client=client, config=example_app_config)
        self.sensor_id = sensor_id
        self.config = example_app_config

        self.detection_retention_duration = int(
            example_app_config.detection_retention_duration * example_app_config.frame_rate
        )

        self.started = False

    def start(
        self, recorder: Optional[a121.Recorder] = None, _algo_group: Optional[h5py.Group] = None
    ) -> None:
        if self.started:
            raise RuntimeError("Already started")

        self._reinitialize_state_variables()

        sensor_config = self._get_sensor_config(self.config)
        self.session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )

        metadata = self.client.setup_session(self.session_config)
        assert isinstance(metadata, a121.Metadata)

        self.processors = self._create_processors(
            sensor_config=sensor_config, example_app_config=self.config, metadata=metadata
        )

        if _algo_group is None and isinstance(recorder, a121.H5Recorder):
            _algo_group = recorder.require_algo_group("faucet")

        if _algo_group is not None:
            _record_algo_data(_algo_group, self.sensor_id, self.config)

        if recorder is not None:
            self.client.attach_recorder(recorder)

        self.client.start_session()

        self.started = True

    def _reinitialize_state_variables(self) -> None:
        num_points_history = int(self.HISTORY_LENGTH_S * self.config.frame_rate)
        self.history = np.zeros(num_points_history)
        self.history_time = np.linspace(-self.HISTORY_LENGTH_S, 0, num_points_history)
        self.has_detected = False
        self.update_index = 0

    def get_next(self) -> ExampleAppResult:
        if not self.started:
            raise RuntimeError("Not started")

        result = self.client.get_next()
        assert isinstance(result, a121.Result)

        assert self.processors is not None
        processor_result = [processor.process(result) for processor in self.processors]

        return self._process_processor_result(processor_result, self.config)

    def _process_processor_result(
        self, results: list[ProcessorResult], config: ExampleAppConfig
    ) -> ExampleAppResult:
        # Determine max presence score and if presence has been detected.
        presence_detected = False
        max_presence_scrore = results[0].inter_presence_score
        for result in results:
            if max_presence_scrore < result.inter_presence_score:
                max_presence_scrore = result.inter_presence_score

            if max_presence_scrore < result.intra_presence_score:
                max_presence_scrore = result.intra_presence_score

            if result.presence_detected:
                presence_detected = True

        # Determine detections state.
        detection_state = DetectionState.NO_DETECTION

        if presence_detected:
            self.has_detected = True
            self.update_index_at_detection = self.update_index
            detection_state = DetectionState.DETECTION

        if (
            self.has_detected
            and detection_state is not DetectionState.DETECTION
            and (self.update_index - self.update_index_at_detection)
            < self.detection_retention_duration
        ):
            detection_state = DetectionState.RETENTION

        # Prepare extra result(used for plotting).
        self.history = np.roll(self.history, shift=-1, axis=0)
        self.history[-1] = max_presence_scrore
        extra_result = ExtraResult(
            history=self.history,
            history_time=self.history_time,
            threshold=config.threshold,
        )

        self.update_index += 1

        return ExampleAppResult(detection_state=detection_state, extra_result=extra_result)

    def stop_detector(self) -> Any:
        if not self.started:
            raise RuntimeError("Already stopped")

        self.client.stop_session()
        self.started = False

        return None

    def stop_recorder(self) -> Any:
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        return recorder_result

    def stop(self) -> Any:
        self.stop_detector()
        recorder_result = self.stop_recorder()

        return recorder_result

    def _create_processors(
        self,
        sensor_config: a121.SensorConfig,
        example_app_config: ExampleAppConfig,
        metadata: a121.Metadata,
    ) -> list[Processor]:
        processors = []
        for subsweep_index, subsweep_config in enumerate(sensor_config.subsweeps):
            processor_sensor_config = a121.SensorConfig(
                start_point=subsweep_config.start_point,
                num_points=subsweep_config.num_points,
                step_length=subsweep_config.step_length,
                profile=subsweep_config.profile,
                frame_rate=sensor_config.frame_rate,
                sweeps_per_frame=sensor_config.sweeps_per_frame,
            )

            processors.append(
                Processor(
                    sensor_config=processor_sensor_config,
                    processor_config=self._get_processor_config(example_app_config),
                    context=ProcessorContext(estimated_frame_rate=example_app_config.frame_rate),
                    metadata=metadata,
                    subsweep_index=subsweep_index,
                )
            )

        return processors

    @classmethod
    def _get_sensor_config(cls, config: ExampleAppConfig) -> a121.SensorConfig:
        """Convert example app config parameters to sensor config.

        Creates a subsweep configuration with
        - A subsweep between the sensor and the water jet(if it fits).
        - A subsweep between the water jet and the end of the measurement range(if it fits).
        """
        profile = a121.Profile.PROFILE_1
        step_length = 6
        water_jet_half_width = config.water_jet_width / 2.0

        subsweep_configs = []

        # Determine if subsweep between sensor and water jet is feasible
        if ENVELOPE_FWHM_M[profile] < (config.sensor_to_water_distance - water_jet_half_width):
            subsweep_config = a121.SubsweepConfig()
            subsweep_config.start_point = int(
                (config.sensor_to_water_distance - ENVELOPE_FWHM_M[profile] - water_jet_half_width)
                / APPROX_BASE_STEP_LENGTH_M
            )
            subsweep_config.num_points = 1
            subsweep_config.step_length = step_length
            subsweep_config.profile = profile
            subsweep_config.hwaas = config.hwaas
            subsweep_config.receiver_gain = 4
            subsweep_configs.append(subsweep_config)

        # Determine if subsweep after water jet is required
        if (
            ENVELOPE_FWHM_M[profile] + water_jet_half_width + config.sensor_to_water_distance
            < config.measurement_range_end
        ):
            start_m = (
                config.sensor_to_water_distance + water_jet_half_width + ENVELOPE_FWHM_M[profile]
            )
            start_point = int(start_m / APPROX_BASE_STEP_LENGTH_M)
            num_points = max(
                1,
                int(
                    (config.measurement_range_end - start_m)
                    / (step_length * APPROX_BASE_STEP_LENGTH_M)
                ),
            )

            subsweep_config = a121.SubsweepConfig()
            subsweep_config.start_point = start_point
            subsweep_config.num_points = num_points
            subsweep_config.step_length = step_length
            subsweep_config.profile = profile
            subsweep_config.hwaas = config.hwaas
            subsweep_config.receiver_gain = 4
            subsweep_configs.append(subsweep_config)

        sensor_config = a121.SensorConfig(subsweeps=subsweep_configs)
        sensor_config.frame_rate = config.frame_rate
        sensor_config.sweep_rate = config.sweeps_rate
        sensor_config.sweeps_per_frame = config.sweeps_per_frame
        sensor_config.inter_frame_idle_state = a121.IdleState.SLEEP
        sensor_config._inter_sweep_idle_state = a121.IdleState.SLEEP

        return sensor_config

    @classmethod
    def _get_processor_config(cls, config: ExampleAppConfig) -> ProcessorConfig:
        return ProcessorConfig(
            intra_enable=True,
            inter_enable=True,
            intra_detection_threshold=config.threshold,
            inter_detection_threshold=config.threshold,
            inter_phase_boost=False,
            phase_adaptivity_tc=5,
            inter_frame_presence_timeout=None,
            inter_frame_fast_cutoff=3.0,
            inter_frame_slow_cutoff=0.8,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=config.filter_time_const,
            intra_frame_time_const=0.15,
            intra_output_time_const=config.filter_time_const,
        )


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    config: ExampleAppConfig,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "example_app_config", config.to_json())


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, ExampleAppConfig]:
    sensor_id = int(algo_group["sensor_id"][()])
    config = ExampleAppConfig.from_json(algo_group["example_app_config"][()])
    return sensor_id, config



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from acconeer.exptool import a121

from ._processor import ProcessorConfig


def _get_base_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        start_point=80,
        num_points=1,
        step_length=1,
        profile=a121.Profile.PROFILE_3,
        receiver_gain=10,
        hwaas=16,
        sweep_rate=2000,
        inter_frame_idle_state=a121.IdleState.READY,
        inter_sweep_idle_state=a121.IdleState.READY,
    )


def get_low_frequency_sensor_config() -> a121.SensorConfig:
    # Sensor config for running low frequency(low sweep rate) in continuous sweep mode
    sensor_config = _get_base_config()
    sensor_config.continuous_sweep_mode = True
    sensor_config.double_buffering = True
    sensor_config.sweeps_per_frame = 50

    return sensor_config


def get_low_frequency_processor_config() -> ProcessorConfig:
    # Processor config for running low frequency(low sweep rate) in continuous sweep mode

    return ProcessorConfig()


def get_high_frequency_sensor_config() -> a121.SensorConfig:
    # Sensor config for running frame mode
    sensor_config = _get_base_config()
    sensor_config.continuous_sweep_mode = False
    sensor_config.double_buffering = False
    sensor_config.sweeps_per_frame = 2048
    sensor_config.sweep_rate = 10000

    return sensor_config


def get_high_frequency_processor_config() -> ProcessorConfig:
    # Processor config for running frame mode

    return ProcessorConfig(lp_coeff=0.5)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from ._configs import (
    get_high_frequency_processor_config,
    get_high_frequency_sensor_config,
    get_low_frequency_processor_config,
    get_low_frequency_sensor_config,
)
from ._processor import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorExtraResult,
    ProcessorResult,
    ReportedDisplacement,
    _load_algo_data,
)



# Copyright (c) Acconeer AB, 2023
# All rights reserved
"""
This is the configuration file for "doit" (https://pydoit.org/)

$ doit                  # runs all tasks
$ doit -n 2             # runs all tasks, while utilizing 2 processes
$ doit list             # lists all tasks
$ doit list --all       # lists all tasks, including subtasks
$ doit <task1> <task2>  # runs task1 & task2
"""

import doit


PYTHON_VERSIONS = ("3.8", "3.9", "3.10", "3.11", "3.12")
GENERATIONS = ("a121", "a111")

CLI_ARGS = {"port_strategy": doit.get_var("port_strategy", "unique")}


def task_integration_test():
    """Integration test against the mock server on different Python versions"""
    for py_ver in PYTHON_VERSIONS:
        for gen in GENERATIONS:
            if CLI_ARGS["port_strategy"] == "unique":
                port = 6000 + (int(py_ver[2:]) * int(gen[1:]))
                yield {
                    "name": f"{py_ver}-{gen}",
                    "actions": [f"tests/run-{gen}-mock-integration-tests.sh {py_ver} {port}"],
                }
            else:
                yield {
                    "name": f"{py_ver}-{gen}",
                    "actions": [f"tests/run-{gen}-mock-integration-tests.sh {py_ver}"],
                }


def task_test():
    """Unit & integration test session (nox) on different Python versions"""
    for py_ver in PYTHON_VERSIONS:
        yield {
            "name": py_ver,
            "actions": [
                f'nox -s "test(python={py_ver!r})" -- '
                + "--test-groups unit integration app doctest "
                + "--editable"
            ],
        }



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from . import builtin_persistors, core, optimizing_persistors
from .api import deserialize, register_json_presentable, register_persistor, serialize



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Optional, Type

import numpy as np

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._plugins import (
    ProcessorBackendPluginBase,
    ProcessorBackendPluginSharedState,
    ProcessorPluginPreset,
    ProcessorViewPluginBase,
    SetupMessage,
)
from acconeer.exptool.a121.algo._utils import APPROX_BASE_STEP_LENGTH_M
from acconeer.exptool.a121.algo.vibration import (
    Processor,
    ProcessorConfig,
    ProcessorResult,
    ReportedDisplacement,
    get_high_frequency_processor_config,
    get_high_frequency_sensor_config,
    get_low_frequency_processor_config,
    get_low_frequency_sensor_config,
)
from acconeer.exptool.app.new import (
    AppModel,
    Message,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    pidgets,
)


log = logging.getLogger(__name__)


class PluginPresetId(Enum):
    LOW_FREQ = auto()
    HIGH_FREQ = auto()


class BackendPlugin(ProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):
    PLUGIN_PRESETS = {
        PluginPresetId.LOW_FREQ.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_low_frequency_sensor_config()),
            processor_config=get_low_frequency_processor_config(),
        ),
        PluginPresetId.HIGH_FREQ.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_high_frequency_sensor_config()),
            processor_config=get_high_frequency_processor_config(),
        ),
    }

    @classmethod
    def get_processor(cls, state: ProcessorBackendPluginSharedState[ProcessorConfig]) -> Processor:
        if state.metadata is None:
            raise RuntimeError("metadata is None")

        if isinstance(state.metadata, list):
            raise RuntimeError("metadata is unexpectedly extended")

        return Processor(
            sensor_config=state.session_config.sensor_config,
            processor_config=state.processor_config,
            metadata=state.metadata,
        )

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_low_frequency_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        # Note: Incomplete mapping
        return {
            "time_series_length": pidgets.IntPidgetFactory(
                name_label_text="Time series length:",
                limits=(0, None),
            ),
            "lp_coeff": pidgets.FloatSliderPidgetFactory(
                name_label_text="Time filtering coefficient:",
                suffix="",
                limits=(0, 1),
                decimals=2,
            ),
            "threshold_margin": pidgets.FloatSliderPidgetFactory(
                name_label_text="Threshold margin:",
                suffix="um",
                limits=(0, 100),
                decimals=1,
            ),
            "amplitude_threshold": pidgets.FloatPidgetFactory(
                name_label_text="Amplitude threshold:",
                decimals=0,
                limits=(0, None),
            ),
            "reported_displacement_mode": pidgets.EnumPidgetFactory(
                name_label_text="Displacement mode:",
                enum_type=ReportedDisplacement,
                label_mapping={
                    ReportedDisplacement.AMPLITUDE: "Amplitude",
                    ReportedDisplacement.PEAK2PEAK: "Peak to peak",
                },
            ),
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[ProcessorResult] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            if isinstance(message.metadata, list):
                raise RuntimeError("Metadata is unexpectedly extended")

            self.setup(
                metadata=message.metadata,
                sensor_config=message.session_config.sensor_config,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(processor_result=self._plot_job)
        finally:
            self._plot_job = None

    def setup(self, metadata: a121.Metadata, sensor_config: a121.SensorConfig) -> None:
        self.plot_layout.clear()

        self.meas_dist_m = sensor_config.start_point * APPROX_BASE_STEP_LENGTH_M

        pen_blue = et.utils.pg_pen_cycler(0)
        pen_orange = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(0)
        brush_dot = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw_blue = dict(pen=pen_blue, **symbol_kw)
        feat_kw_orange = dict(pen=pen_orange)
        symbol_dot_kw = dict(symbol="o", symbolSize=10, symbolBrush=brush_dot, symbolPen="k")

        # presence plot
        self.object_detection_plot = pg.PlotItem()
        self.object_detection_plot.setMenuEnabled(False)
        self.object_detection_plot.showGrid(x=False, y=True)
        self.object_detection_plot.setLabel("left", "Max amplitude")
        self.object_detection_plot.setLabel("bottom", "Distance (m)")
        self.object_detection_plot.setXRange(self.meas_dist_m - 0.001, self.meas_dist_m + 0.001)
        self.presence_curve = self.object_detection_plot.plot(
            **dict(pen=pen_blue, **symbol_dot_kw)
        )

        self.presence_threshold = pg.InfiniteLine(pen=pen_blue, angle=0)
        self.object_detection_plot.addItem(self.presence_threshold)
        self.presence_threshold.show()

        self.smooth_max_presence = et.utils.SmoothMax(tau_decay=10.0)

        # sweep and threshold plot
        self.time_series_plot = pg.PlotItem()
        self.time_series_plot.setMenuEnabled(False)
        self.time_series_plot.showGrid(x=True, y=True)
        self.time_series_plot.setLabel("left", "Displacement (<font>&mu;</font>m)")
        self.time_series_plot.setLabel("bottom", "History")
        self.time_series_curve = self.time_series_plot.plot(**feat_kw_blue)

        self.time_series_plot.setYRange(-1000, 1000)
        self.time_series_plot.setXRange(0, 1024)

        self.text_item_time_series = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.text_item_time_series.hide()
        self.time_series_plot.addItem(self.text_item_time_series)

        sublayout = self.plot_layout.addLayout(row=0, col=0)
        sublayout.layout.setColumnStretchFactor(1, 5)
        sublayout.addItem(self.object_detection_plot, row=0, col=0)
        sublayout.addItem(self.time_series_plot, row=0, col=1)

        self.smooth_lim_time_series = et.utils.SmoothLimits(tau_decay=0.5, tau_grow=0.1)

        self.fft_plot = self.plot_layout.addPlot(col=0, row=1)
        self.fft_plot.setMenuEnabled(False)
        self.fft_plot.showGrid(x=True, y=True)
        self.fft_plot.setLabel("left", "Displacement (<font>&mu;</font>m)")
        self.fft_plot.setLabel("bottom", "Frequency (Hz)")
        self.fft_plot.setLogMode(False, True)
        self.fft_plot.addItem(pg.PlotDataItem())
        self.fft_curve = [
            self.fft_plot.plot(**feat_kw_blue),
            self.fft_plot.plot(**feat_kw_orange),
            self.fft_plot.plot(**dict(pen=pen_blue, **symbol_dot_kw)),
        ]

        self.text_item_fft = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.text_item_fft.hide()
        self.fft_plot.addItem(self.text_item_fft)

        self.smooth_max_fft = et.utils.SmoothMax()

    def draw_plot_job(self, processor_result: ProcessorResult) -> None:
        # Extra result
        time_series = processor_result.extra_result.zm_time_series
        lp_displacements_threshold = processor_result.extra_result.lp_displacements_threshold
        amplitude_threshold = processor_result.extra_result.amplitude_threshold

        # Processor result
        lp_displacements = processor_result.lp_displacements
        lp_displacements_freqs = processor_result.lp_displacements_freqs
        max_amplitude = processor_result.max_sweep_amplitude
        max_displacement = processor_result.max_displacement
        max_displacement_freq = processor_result.max_displacement_freq
        time_series_std = processor_result.time_series_std

        # Plot object presence metric
        self.presence_curve.setData([self.meas_dist_m], [max_amplitude])
        self.presence_threshold.setValue(amplitude_threshold)
        lim = self.smooth_max_presence.update(max_amplitude)
        self.object_detection_plot.setYRange(0, max(1000.0, lim))

        # Plot time series
        if time_series is not None and amplitude_threshold < max_amplitude:
            assert time_series_std is not None
            lim = self.smooth_lim_time_series.update(time_series)
            self.time_series_plot.setYRange(lim[0], lim[1])
            self.time_series_plot.setXRange(0, time_series.shape[0])

            self.text_item_time_series.setPos(time_series.size / 2, lim[1] * 0.95)
            time_series_std_str = "{:.0f}".format(time_series_std)
            html_format = (
                '<div style="text-align: center">'
                '<span style="color: #FFFFFF;font-size:15pt;">'
                "{}</span></div>".format("STD : " + time_series_std_str + "<font>&mu;</font>m")
            )
            self.text_item_time_series.setHtml(html_format)
            self.text_item_time_series.hide()  # do not display std(for now)
            self.time_series_curve.setData(time_series)

        # Plot spectrum
        if lp_displacements is not None:
            assert time_series is not None
            assert lp_displacements is not None

            self.fft_curve[0].setData(lp_displacements_freqs, lp_displacements)
            self.fft_curve[1].setData(lp_displacements_freqs, lp_displacements_threshold)
            lim = self.smooth_max_fft.update(np.max(lp_displacements))
            self.fft_plot.setYRange(-1, np.log10(lim))

            if max_displacement_freq is not None and max_displacement is not None:
                self.fft_curve[2].setData([max_displacement_freq], [max_displacement])

                # Place text box centered at the top of the plotting window
                self.text_item_fft.setPos(max(lp_displacements_freqs) / 2, np.log10(lim) * 0.95)
                max_displacement_str = "{:.0f}".format(max_displacement)
                max_displacement_freq_str = "{:.1f}".format(max_displacement_freq)
                html_format = (
                    '<div style="text-align: center">'
                    '<span style="color: #FFFFFF;font-size:15pt;">'
                    "{}</span></div>".format(
                        "Frequency: "
                        + max_displacement_freq_str
                        + " Hz"
                        + "<br>"
                        + "Displacement: "
                        + max_displacement_str
                        + "<font>&mu;</font>m"
                    )
                )
                self.text_item_fft.setHtml(html_format)
                self.text_item_fft.show()
            else:
                self.fft_curve[2].setData([], [])
                self.text_item_fft.hide()


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


VIBRATION_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="vibration",
    title="Vibration measurement",
    docs_link="https://docs.acconeer.com/en/latest/exploration_tool/algo/a121/examples/vibration.html",
    description="Quantify the frequency content of vibrating object.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(name="Low frequency", preset_id=PluginPresetId.LOW_FREQ),
        PluginPresetBase(name="High frequency", preset_id=PluginPresetId.HIGH_FREQ),
    ],
    default_preset_id=PluginPresetId.LOW_FREQ,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401
from PySide6.QtGui import QFont

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._utils import get_distances_m
from acconeer.exptool.a121.algo.breathing import AppState, RefApp
from acconeer.exptool.a121.algo.breathing._ref_app import (
    BreathingProcessorConfig,
    RefAppConfig,
    get_sensor_config,
)
from acconeer.exptool.a121.algo.presence import ProcessorConfig as PresenceProcessorConfig


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    # Setup the configurations
    # Detailed at https://docs.acconeer.com/en/latest/exploration_tool/algo/a121/ref_apps/breathing.html

    # Sensor selections
    sensor = 1

    # Ref App Configurations
    breathing_processor_config = BreathingProcessorConfig(
        lowest_breathing_rate=6,
        highest_breathing_rate=60,
        time_series_length_s=20,
    )

    # Presence Configurations
    presence_config = PresenceProcessorConfig(
        intra_detection_threshold=4,
        intra_frame_time_const=0.15,
        inter_frame_fast_cutoff=20,
        inter_frame_slow_cutoff=0.2,
        inter_frame_deviation_time_const=0.5,
    )

    # Breathing Configurations
    ref_app_config = RefAppConfig(
        use_presence_processor=True,
        num_distances_to_analyze=3,
        distance_determination_duration=5,
        breathing_config=breathing_processor_config,
        presence_config=presence_config,
    )

    # End setup configurations

    # Preparation for client
    sensor_config = get_sensor_config(ref_app_config=ref_app_config)
    client = a121.Client.open(**a121.get_client_args(args))
    metadata = client.setup_session(sensor_config)

    # Preparation for reference application processor
    ref_app = RefApp(client=client, sensor_id=sensor, ref_app_config=ref_app_config)
    ref_app.start()

    pg_updater = PGUpdater(sensor_config, ref_app_config, metadata)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        processed_data = ref_app.get_next()
        try:
            pg_process.put_data(processed_data)
        except et.PGProccessDiedException:
            break

    ref_app.stop()
    print("Disconnecting...")
    client.close()


class PGUpdater:
    def __init__(
        self,
        sensor_config: a121.SensorConfig,
        ref_app_config: RefAppConfig,
        metadata: a121.Metadata,
    ):
        self.distances = get_distances_m(sensor_config, metadata)
        self.use_presence_processor = ref_app_config.use_presence_processor

    def setup(self, win):
        # Define pens and font.
        blue_color = et.utils.color_cycler(0)
        orange_color = et.utils.color_cycler(1)
        brush = et.utils.pg_brush_cycler(0)
        self.blue = dict(
            pen=pg.mkPen(blue_color, width=2),
            symbol="o",
            symbolSize=1,
            symbolBrush=brush,
            symbolPen="k",
        )
        self.orange = dict(
            pen=pg.mkPen(orange_color, width=2),
            symbol="o",
            symbolSize=1,
            symbolBrush=brush,
            symbolPen="k",
        )
        self.blue_transparent_pen = pg.mkPen(f"{blue_color}50", width=2)
        self.orange_transparent_pen = pg.mkPen(f"{orange_color}50", width=2)

        brush_dot = et.utils.pg_brush_cycler(1)
        symbol_dot_kw = dict(symbol="o", symbolSize=10, symbolBrush=brush_dot, symbolPen="k")

        font = QFont()
        font.setPixelSize(16)

        # Presence plot.
        self.presence_plot = win.addPlot(row=0, col=0)
        self.presence_plot.setMenuEnabled(False)
        self.presence_plot.showGrid(x=True, y=True)
        self.presence_plot.addLegend()
        self.presence_plot.setLabel("left", "Presence score")
        self.presence_plot.setLabel("bottom", "Distance (m)")
        self.presence_plot.addItem(pg.PlotDataItem())
        self.presence_plot_curve = []
        self.presence_plot_curve.append(self.presence_plot.plot(**self.blue))
        self.presence_plot_curve.append(self.presence_plot.plot(**self.orange))
        self.presence_plot_curve.append(self.presence_plot.plot(**self.blue))
        self.presence_plot_curve.append(self.presence_plot.plot(**self.orange))

        self.presence_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        self.presence_plot_legend.setParentItem(self.presence_plot)
        self.presence_plot_legend.addItem(self.presence_plot_curve[2], "Slow motion")
        self.presence_plot_legend.addItem(self.presence_plot_curve[3], "Fast motion")
        self.presence_plot_legend.show()

        self.presence_smoot_max = et.utils.SmoothMax()

        self.presence_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.presence_text_item.setFont(font)
        self.presence_text_item.show()
        self.presence_plot.addItem(self.presence_text_item)

        # Time series plot.
        self.time_series_plot = win.addPlot(row=1, col=0)
        self.time_series_plot.setMenuEnabled(False)
        self.time_series_plot.showGrid(x=True, y=True)
        self.time_series_plot.addLegend()
        self.time_series_plot.setLabel("left", "Displacement")
        self.time_series_plot.setLabel("bottom", "Time (s)")
        self.time_series_plot.addItem(pg.PlotDataItem())
        self.time_series_curve = self.time_series_plot.plot(**self.blue)

        self.time_series_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.time_series_text_item.setFont(font)
        self.time_series_text_item.show()
        self.time_series_plot.addItem(self.time_series_text_item)

        # Breathing psd plot.
        self.breathing_psd_plot = win.addPlot(row=2, col=0)
        self.breathing_psd_plot.setMenuEnabled(False)
        self.breathing_psd_plot.showGrid(x=True, y=True)
        self.breathing_psd_plot.addLegend()
        self.breathing_psd_plot.setLabel("left", "PSD")
        self.breathing_psd_plot.setLabel("bottom", "Breathing rate (Hz)")
        self.breathing_psd_plot.addItem(pg.PlotDataItem())
        self.breathing_psd_curve = self.breathing_psd_plot.plot(**self.blue)

        self.psd_smoothing = et.utils.SmoothMax()

        # Breathing rate plot.
        self.breathing_rate_plot = win.addPlot(row=3, col=0)
        self.breathing_rate_plot.setMenuEnabled(False)
        self.breathing_rate_plot.showGrid(x=True, y=True)
        self.breathing_rate_plot.addLegend()
        self.breathing_rate_plot.setLabel("left", "Breaths per minute")
        self.breathing_rate_plot.setLabel("bottom", "Time (s)")
        self.breathing_rate_plot.addItem(pg.PlotDataItem())
        self.breathing_rate_curves = []
        self.breathing_rate_curves.append(self.breathing_rate_plot.plot(**self.blue))
        self.breathing_rate_curves.append(
            self.breathing_rate_plot.plot(**dict(pen=None, **symbol_dot_kw))
        )
        self.smooth_breathing_rate = et.utils.SmoothLimits()

        self.breathing_rate_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        self.breathing_rate_plot_legend.setParentItem(self.breathing_rate_plot)
        self.breathing_rate_plot_legend.addItem(self.breathing_rate_curves[0], "Breathing rate")
        self.breathing_rate_plot_legend.addItem(
            self.breathing_rate_curves[1], "Breathing rate(embedded output)"
        )
        self.breathing_rate_plot_legend.hide()

        self.breathing_rate_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.breathing_rate_text_item.setFont(font)
        self.breathing_rate_text_item.hide()
        self.breathing_rate_plot.addItem(self.breathing_rate_text_item)

    def update(self, ref_app_result):
        app_state = ref_app_result.app_state

        max_ampl = max(
            np.max(ref_app_result.presence_result.inter),
            np.max(ref_app_result.presence_result.intra),
        )
        lim = self.presence_smoot_max.update(max_ampl)
        self.presence_plot.setYRange(0, lim)

        if ref_app_result.distances_being_analyzed is None:
            self.presence_plot_curve[0].setData(
                self.distances, ref_app_result.presence_result.inter, **self.blue
            )
            self.presence_plot_curve[1].setData(
                self.distances, ref_app_result.presence_result.intra, **self.orange
            )
            self.presence_plot_curve[2].setData([], [])
            self.presence_plot_curve[3].setData([], [])
        else:
            start = ref_app_result.distances_being_analyzed[0]
            end = ref_app_result.distances_being_analyzed[1]
            s = slice(start, end)
            distance_slice = self.distances[s]
            self.presence_plot_curve[0].setData(
                self.distances,
                ref_app_result.presence_result.inter,
                pen=self.blue_transparent_pen,
            )
            self.presence_plot_curve[1].setData(
                self.distances,
                ref_app_result.presence_result.intra,
                pen=self.orange_transparent_pen,
            )
            self.presence_plot_curve[2].setData(
                distance_slice, ref_app_result.presence_result.inter[s]
            )
            self.presence_plot_curve[3].setData(
                distance_slice, ref_app_result.presence_result.intra[s]
            )

        if ref_app_result.breathing_result is not None:
            breathing_result = ref_app_result.breathing_result.extra_result
            breathing_motion = breathing_result.breathing_motion
            psd = breathing_result.psd
            frequencies = breathing_result.frequencies
            time_vector = breathing_result.time_vector
            all_breathing_rate_history = breathing_result.all_breathing_rate_history
            breathing_rate_history = breathing_result.breathing_rate_history

            self.time_series_curve.setData(
                time_vector[-breathing_motion.shape[0] :], breathing_motion
            )
            y = np.max(np.abs(breathing_motion)) * 1.05
            self.time_series_plot.setYRange(-y, y)
            self.time_series_plot.setXRange(
                time_vector[-breathing_motion.shape[0]], max(time_vector)
            )

            if not np.all(np.isnan(all_breathing_rate_history)):
                ylim = self.psd_smoothing.update(psd)
                self.breathing_psd_curve.setData(frequencies, psd)
                self.breathing_psd_plot.setYRange(0, ylim)
                self.breathing_psd_plot.setXRange(0, 2)

                self.breathing_rate_curves[0].setData(time_vector, all_breathing_rate_history)
                lims = self.smooth_breathing_rate.update(all_breathing_rate_history)
                self.breathing_rate_plot.setYRange(lims[0] - 3, lims[1] + 3)

                self.breathing_rate_plot_legend.show()

            if not np.all(np.isnan(breathing_rate_history)):
                self.breathing_rate_curves[1].setData(time_vector, breathing_rate_history)

            if not np.isnan(breathing_rate_history[-1]):
                self.displayed_breathing_rate = "{:.1f}".format(breathing_rate_history[-1])
                self.breathing_rate_text_item.show()

        else:
            self.time_series_plot.setYRange(0, 1)
            self.time_series_plot.setXRange(0, 1)
            self.time_series_curve.setData([], [])
            self.breathing_psd_curve.setData([], [])
            self.breathing_rate_curves[0].setData([], [])
            self.breathing_rate_curves[1].setData([], [])
            self.displayed_breathing_rate = None
            self.breathing_rate_text_item.hide()

        # Set text in text boxes according to app state.

        # Presence text
        if app_state == AppState.NO_PRESENCE_DETECTED:
            presence_text = "No presence detected"
        elif app_state == AppState.DETERMINE_DISTANCE_ESTIMATE:
            presence_text = "Determining distance with presence"
        elif app_state == AppState.ESTIMATE_BREATHING_RATE:
            start_m = "{:.2f}".format(distance_slice[0])
            end_m = "{:.2f}".format(distance_slice[-1])
            if self.use_presence_processor:
                presence_text = (
                    "Presence detected in the range " + start_m + " - " + end_m + " (m)"
                )
            else:
                presence_text = "Presence distance detection disabled"
        elif app_state == AppState.INTRA_PRESENCE_DETECTED:
            presence_text = "Large motion detected"
        else:
            presence_text = ""

        text_y_pos = self.presence_plot.getAxis("left").range[1] * 0.95
        text_x_pos = (
            self.presence_plot.getAxis("bottom").range[1]
            + self.presence_plot.getAxis("bottom").range[0]
        ) / 2.0
        self.presence_text_item.setPos(text_x_pos, text_y_pos)
        self.presence_text_item.setHtml(presence_text)

        # Breathing text
        if app_state == AppState.ESTIMATE_BREATHING_RATE:
            if (
                ref_app_result.breathing_result is not None
                and ref_app_result.breathing_result.breathing_rate is None
            ):
                time_series_text = "Initializing breathing detection"
            elif self.displayed_breathing_rate is not None:
                time_series_text = "Breathing rate: " + self.displayed_breathing_rate + " bpm"
        else:
            time_series_text = "Waiting for distance"

        text_y_pos = self.time_series_plot.getAxis("left").range[1] * 0.95
        text_x_pos = (
            self.time_series_plot.getAxis("bottom").range[1]
            + self.time_series_plot.getAxis("bottom").range[0]
        ) / 2.0
        self.time_series_text_item.setPos(text_x_pos, text_y_pos)
        self.time_series_text_item.setHtml(time_series_text)

        if self.displayed_breathing_rate is not None:
            text_y_pos = self.breathing_rate_plot.getAxis("left").range[1] * 0.95
            text_x_pos = time_vector[0]

            self.breathing_rate_text_item.setPos(text_x_pos, text_y_pos)
            self.breathing_rate_text_item.setHtml(self.displayed_breathing_rate + " bpm")


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import numpy as np

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.touchless_button import (
    MeasurementType,
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_close_sensor_config,
)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client.open(**a121.get_client_args(args))

    processor_config = ProcessorConfig()

    sensor_config = get_close_sensor_config()

    metadata = client.setup_session(sensor_config)
    client.start_session()

    processor = Processor(
        sensor_config=sensor_config,
        metadata=metadata,
        processor_config=processor_config,
    )

    pg_updater = PGUpdater(sensor_config=sensor_config, processor_config=processor_config)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        result = client.get_next()
        processor_result = processor.process(result)
        try:
            pg_process.put_data(processor_result)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    pg_process.close()
    client.stop_session()
    client.close()


class PGUpdater:
    def __init__(self, sensor_config, processor_config):
        self.detection_history = np.full((2, 100), np.NaN)
        self.sensor_config = sensor_config
        self.processor_config = processor_config

    def setup(self, win):
        self.detection_history_plot = self._create_detection_plot(win)

        self.detection_history_curve_close = self.detection_history_plot.plot(
            pen=et.utils.pg_pen_cycler(1, width=5)
        )
        self.detection_history_curve_far = self.detection_history_plot.plot(
            pen=et.utils.pg_pen_cycler(0, width=5)
        )

        close_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Close detection")
        )
        far_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Far detection")
        )
        self.close_text_item = pg.TextItem(
            html=close_html,
            fill=pg.mkColor(0xFF, 0x7F, 0x0E),
            anchor=(0.5, 0),
        )
        self.far_text_item = pg.TextItem(
            html=far_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4),
            anchor=(0.5, 0),
        )
        pos_left = (100 / 3, 1.8)
        pos_right = (2 * 100 / 3, 1.8)
        self.close_text_item.setPos(*pos_left)
        self.far_text_item.setPos(*pos_right)
        self.detection_history_plot.addItem(self.close_text_item)
        self.detection_history_plot.addItem(self.far_text_item)
        self.close_text_item.hide()
        self.far_text_item.hide()

        self.detection_history = np.full((2, 100), np.NaN)

        self.score_history_plot = self._create_score_plot(win)
        score_plot_legend = self.score_history_plot.legend
        self.score_smooth_max = et.utils.SmoothMax()

        self.threshold_history_curve_close = self.score_history_plot.plot(
            pen=et.utils.pg_pen_cycler(1, width=2.5, style="--"),
        )
        self.threshold_history_curve_far = self.score_history_plot.plot(
            pen=et.utils.pg_pen_cycler(0, width=2.5, style="--"),
        )

        self.threshold_history = np.full((2, 100), np.NaN)

        cycle_index = 2  # To not have same colors as thresholds
        if self.processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            measurement_type = "Close"
            score_plot_legend.addItem(self.threshold_history_curve_close, "Close range threshold")
        elif self.processor_config.measurement_type == MeasurementType.FAR_RANGE:
            measurement_type = "Far"
            score_plot_legend.addItem(self.threshold_history_curve_far, "Far range threshold")

        if self.processor_config.measurement_type != MeasurementType.CLOSE_AND_FAR_RANGE:
            score_history = np.full((self.sensor_config.subsweep.num_points, 100), np.NaN)
            score_history_curves = np.empty(
                (self.sensor_config.subsweep.num_points,), dtype=object
            )
            for i in range(self.sensor_config.subsweep.num_points):
                score_history_curves[i] = pg.ScatterPlotItem(
                    brush=et.utils.pg_brush_cycler(cycle_index),
                    name=f"{measurement_type} range, point {i}",
                )
                self.score_history_plot.addItem(score_history_curves[i])
                cycle_index += 1

            if self.processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
                self.score_history_close = score_history
                self.score_history_curves_close = score_history_curves
                self.score_history_far = None
                self.score_history_curves_far = None
            elif self.processor_config.measurement_type == MeasurementType.FAR_RANGE:
                self.score_history_close = None
                self.score_history_curves_close = None
                self.score_history_far = score_history
                self.score_history_curves_far = score_history_curves

        elif self.processor_config.measurement_type == MeasurementType.CLOSE_AND_FAR_RANGE:
            score_plot_legend.addItem(self.threshold_history_curve_close, "Close range threshold")
            score_plot_legend.addItem(self.threshold_history_curve_far, "Far range threshold")
            self.score_history_close = np.full(
                (self.sensor_config.subsweeps[0].num_points, 100), np.NaN
            )
            self.score_history_far = np.full(
                (self.sensor_config.subsweeps[1].num_points, 100), np.NaN
            )
            self.score_history_curves_close = np.empty(
                (self.sensor_config.subsweeps[0].num_points,), dtype=object
            )
            self.score_history_curves_far = np.empty(
                (self.sensor_config.subsweeps[1].num_points,), dtype=object
            )

            range_labels = ["Close", "Far"]
            for n, subsweep in enumerate(self.sensor_config.subsweeps):
                measurement_type = range_labels[n]
                score_history_curve_list = [
                    self.score_history_curves_close,
                    self.score_history_curves_far,
                ]
                for i in range(subsweep.num_points):
                    score_history_curve_list[n][i] = pg.ScatterPlotItem(
                        brush=et.utils.pg_brush_cycler(cycle_index),
                        name=f"{measurement_type} range, point {i}",
                    )
                    self.score_history_plot.addItem(score_history_curve_list[n][i])
                    cycle_index += 1
            self.score_history_curves_close = score_history_curve_list[0]
            self.score_history_curves_far = score_history_curve_list[1]

    def update(self, processor_result: ProcessorResult):
        def is_none_or_detection(x):
            return x.detection if x is not None else None

        detection = np.array(
            [
                is_none_or_detection(processor_result.close),
                is_none_or_detection(processor_result.far),
            ]
        )
        self.detection_history = np.roll(self.detection_history, -1, axis=1)
        self.detection_history[:, -1] = detection

        self.detection_history_curve_close.setData(self.detection_history[0])
        self.detection_history_curve_far.setData(self.detection_history[1])

        if processor_result.close is not None:
            if processor_result.close.detection:
                self.close_text_item.show()
            else:
                self.close_text_item.hide()

        if processor_result.far is not None:
            if processor_result.far.detection:
                self.far_text_item.show()
            else:
                self.far_text_item.hide()

        max_val = 0.0

        def is_none_or_threshold(x):
            return x.threshold if x is not None else None

        threshold = np.array(
            [
                is_none_or_threshold(processor_result.close),
                is_none_or_threshold(processor_result.far),
            ]
        )
        if np.nanmax(np.array(threshold, dtype=float)) > max_val:
            max_val = np.nanmax(np.array(threshold, dtype=float))
        self.threshold_history = np.roll(self.threshold_history, -1, axis=1)
        self.threshold_history[:, -1] = threshold

        self.threshold_history_curve_close.setData(self.threshold_history[0])
        self.threshold_history_curve_far.setData(self.threshold_history[1])

        if self.score_history_close is not None:
            self.score_history_close = np.roll(self.score_history_close, -1, axis=1)
            assert processor_result.close is not None
            # Plot the second highest score
            self.score_history_close[:, -1] = np.sort(processor_result.close.score, axis=0)[-2, :]

            assert self.score_history_curves_close is not None
            for i, curve in enumerate(self.score_history_curves_close):
                # Assign x-values so that setData() doesn't give error when y-values are NaN
                curve.setData(np.arange(0, 100), self.score_history_close[i, :].flatten())

            if np.max(processor_result.close.score) > max_val:
                max_val = np.max(processor_result.close.score)

        if self.score_history_far is not None:
            self.score_history_far = np.roll(self.score_history_far, -1, axis=1)
            assert processor_result.far is not None
            # Plot the second highest score
            self.score_history_far[:, -1] = np.sort(processor_result.far.score, axis=0)[-2, :]

            assert self.score_history_curves_far is not None
            for i, curve in enumerate(self.score_history_curves_far):
                # Assign x-values so that setData() doesn't give error when y-values are NaN
                curve.setData(np.arange(0, 100), self.score_history_far[i, :].flatten())

            if np.max(processor_result.far.score) > max_val:
                max_val = np.max(processor_result.far.score)

        if max_val != 0.0:
            self.score_history_plot.setYRange(0.0, self.score_smooth_max.update(max_val))

    @staticmethod
    def _create_detection_plot(parent: pg.GraphicsLayout) -> pg.PlotItem:
        detection_history_plot = parent.addPlot(row=0, col=0)
        detection_history_plot.setTitle("Detection")
        detection_history_plot.setLabel(axis="bottom", text="Frames")
        detection_history_plot.setMenuEnabled(False)
        detection_history_plot.setMouseEnabled(x=False, y=False)
        detection_history_plot.hideButtons()
        detection_history_plot.showGrid(x=True, y=True, alpha=0.5)
        detection_history_plot.setYRange(-0.1, 1.8)
        return detection_history_plot

    @staticmethod
    def _create_score_plot(parent: pg.GraphicsLayout) -> pg.PlotItem:
        score_history_plot = parent.addPlot(row=1, col=0)
        score_history_plot.setTitle("Detection score")
        score_history_plot.setLabel(axis="bottom", text="Frames")
        score_history_plot.addLegend()
        score_history_plot.setMenuEnabled(False)
        score_history_plot.setMouseEnabled(x=False, y=False)
        score_history_plot.hideButtons()
        score_history_plot.showGrid(x=True, y=True, alpha=0.5)
        return score_history_plot


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._utils import APPROX_BASE_STEP_LENGTH_M
from acconeer.exptool.a121.algo.vibration import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
    get_high_frequency_sensor_config,
)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    sensor_config = get_high_frequency_sensor_config()

    client = a121.Client.open(**a121.get_client_args(args))
    metadata = client.setup_session(sensor_config)

    processor = Processor(
        sensor_config=sensor_config,
        metadata=metadata,
        processor_config=ProcessorConfig(),
        context=ProcessorContext(),
    )
    pg_updater = PGUpdater(sensor_config)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    client.start_session()
    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        result = client.get_next()
        plot_data = processor.process(result)
        try:
            pg_process.put_data(plot_data)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    client.close()


class PGUpdater:
    def __init__(self, sensor_config):
        self.meas_dist_m = sensor_config.start_point * APPROX_BASE_STEP_LENGTH_M
        self.sensor_config = sensor_config

    def setup(self, win):
        self.meas_dist_m = self.sensor_config.start_point * APPROX_BASE_STEP_LENGTH_M

        pen_blue = et.utils.pg_pen_cycler(0)
        pen_orange = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(0)
        brush_dot = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw_blue = dict(pen=pen_blue, **symbol_kw)
        feat_kw_orange = dict(pen=pen_orange)
        symbol_dot_kw = dict(symbol="o", symbolSize=10, symbolBrush=brush_dot, symbolPen="k")

        # presence plot
        self.object_detection_plot = win.addPlot()
        self.object_detection_plot.setMenuEnabled(False)
        self.object_detection_plot.showGrid(x=False, y=True)
        self.object_detection_plot.setLabel("left", "Max amplitude")
        self.object_detection_plot.setLabel("bottom", "Distance (m)")
        self.object_detection_plot.setXRange(self.meas_dist_m - 0.001, self.meas_dist_m + 0.001)
        self.presence_curve = self.object_detection_plot.plot(
            **dict(pen=pen_blue, **symbol_dot_kw)
        )

        self.presence_threshold = pg.InfiniteLine(pen=pen_blue, angle=0)
        self.object_detection_plot.addItem(self.presence_threshold)
        self.presence_threshold.show()

        self.smooth_max_presence = et.utils.SmoothMax(tau_decay=10.0)

        # sweep and threshold plot
        self.time_series_plot = win.addPlot()
        self.time_series_plot.setMenuEnabled(False)
        self.time_series_plot.showGrid(x=True, y=True)
        self.time_series_plot.setLabel("left", "Displacement (<font>&mu;</font>m)")
        self.time_series_plot.setLabel("bottom", "History")
        self.time_series_curve = self.time_series_plot.plot(**feat_kw_blue)

        self.time_series_plot.setYRange(-1000, 1000)
        self.time_series_plot.setXRange(0, 1024)

        self.text_item_time_series = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.text_item_time_series.hide()
        self.time_series_plot.addItem(self.text_item_time_series)

        sublayout = win.addLayout(row=0, col=0)
        sublayout.layout.setColumnStretchFactor(1, 5)
        sublayout.addItem(self.object_detection_plot, row=0, col=0)
        sublayout.addItem(self.time_series_plot, row=0, col=1)

        self.smooth_lim_time_series = et.utils.SmoothLimits(tau_decay=0.5, tau_grow=0.1)

        self.fft_plot = win.addPlot(col=0, row=1)
        self.fft_plot.setMenuEnabled(False)
        self.fft_plot.showGrid(x=True, y=True)
        self.fft_plot.setLabel("left", "Displacement (<font>&mu;</font>m)")
        self.fft_plot.setLabel("bottom", "Frequency (Hz)")
        self.fft_plot.setLogMode(False, True)
        self.fft_plot.addItem(pg.PlotDataItem())
        self.fft_curve = [
            self.fft_plot.plot(**feat_kw_blue),
            self.fft_plot.plot(**feat_kw_orange),
            self.fft_plot.plot(**dict(pen=pen_blue, **symbol_dot_kw)),
        ]

        self.text_item_fft = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.text_item_fft.hide()
        self.fft_plot.addItem(self.text_item_fft)

        self.smooth_max_fft = et.utils.SmoothMax()

    def update(self, processor_result: ProcessorResult) -> None:
        # Extra result
        time_series = processor_result.extra_result.zm_time_series
        lp_displacements_threshold = processor_result.extra_result.lp_displacements_threshold
        amplitude_threshold = processor_result.extra_result.amplitude_threshold

        # Processor result
        lp_displacements = processor_result.lp_displacements
        lp_displacements_freqs = processor_result.lp_displacements_freqs
        max_amplitude = processor_result.max_sweep_amplitude
        max_displacement = processor_result.max_displacement
        max_displacement_freq = processor_result.max_displacement_freq
        time_series_rms = processor_result.time_series_std

        # Plot object presence metric
        self.presence_curve.setData([self.meas_dist_m], [max_amplitude])
        self.presence_threshold.setValue(amplitude_threshold)
        lim = self.smooth_max_presence.update(max_amplitude)
        self.object_detection_plot.setYRange(0, max(1000.0, lim))

        # Plot time series
        if time_series is not None and amplitude_threshold < max_amplitude:
            assert time_series_rms is not None
            lim = self.smooth_lim_time_series.update(time_series)
            self.time_series_plot.setYRange(lim[0], lim[1])
            self.time_series_plot.setXRange(0, time_series.shape[0])

            self.text_item_time_series.setPos(time_series.size / 2, lim[1] * 0.95)
            html_format = (
                '<div style="text-align: center">'
                '<span style="color: #FFFFFF;font-size:15pt;">'
                "{}</span></div>".format("RMS : " + str(int(time_series_rms)))
            )
            self.text_item_time_series.setHtml(html_format)
            self.text_item_time_series.show()
            self.time_series_curve.setData(time_series)

        # Plot spectrum
        if lp_displacements is not None:
            assert time_series is not None
            assert lp_displacements is not None

            self.fft_curve[0].setData(lp_displacements_freqs, lp_displacements)
            self.fft_curve[1].setData(lp_displacements_freqs, lp_displacements_threshold)
            lim = self.smooth_max_fft.update(np.max(lp_displacements))
            self.fft_plot.setYRange(-1, np.log10(lim))

            if max_displacement_freq is not None and max_displacement is not None:
                self.fft_curve[2].setData([max_displacement_freq], [max_displacement])

                # Place text box centered at the top of the plotting window
                self.text_item_fft.setPos(max(lp_displacements_freqs) / 2, np.log10(lim) * 0.95)
                html_format = (
                    '<div style="text-align: center">'
                    '<span style="color: #FFFFFF;font-size:15pt;">'
                    "{}</span></div>".format(
                        "Frequency: "
                        + str(int(max_displacement_freq))
                        + "Hz Displacement: "
                        + str(int(max_displacement))
                        + "<font>&mu;</font>m"
                    )
                )
                self.text_item_fft.setHtml(html_format)
                self.text_item_fft.show()
            else:
                self.fft_curve[2].setData([], [])
                self.text_item_fft.hide()


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.breathing import RefApp
from acconeer.exptool.a121.algo.breathing._ref_app import (
    BreathingProcessorConfig,
    RefAppConfig,
    get_sensor_config,
)
from acconeer.exptool.a121.algo.presence import ProcessorConfig as PresenceProcessorConfig


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    # Setup the configurations
    # Detailed at https://docs.acconeer.com/en/latest/exploration_tool/algo/a121/ref_apps/breathing.html

    # Sensor selections
    sensor = 1

    # Ref App Configurations
    breathing_processor_config = BreathingProcessorConfig(
        lowest_breathing_rate=6,
        highest_breathing_rate=60,
        time_series_length_s=20,
    )

    # Presence Configurations
    presence_config = PresenceProcessorConfig(
        intra_detection_threshold=4,
        intra_frame_time_const=0.15,
        inter_frame_fast_cutoff=20,
        inter_frame_slow_cutoff=0.2,
        inter_frame_deviation_time_const=0.5,
    )

    # Breathing Configurations
    ref_app_config = RefAppConfig(
        use_presence_processor=True,
        num_distances_to_analyze=3,
        distance_determination_duration=5,
        breathing_config=breathing_processor_config,
        presence_config=presence_config,
    )

    # End setup configurations

    # Preparation for client
    sensor_config = get_sensor_config(ref_app_config=ref_app_config)
    client = a121.Client.open(**a121.get_client_args(args))
    client.setup_session(sensor_config)

    # Preparation for reference application processor
    ref_app = RefApp(client=client, sensor_id=sensor, ref_app_config=ref_app_config)
    ref_app.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        processed_data = ref_app.get_next()
        try:
            print("Breathing result " + str(processed_data.breathing_result))
        except et.PGProccessDiedException:
            break

    ref_app.stop()
    print("Disconnecting...")
    client.close()


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from acconeer.exptool.a121.algo.obstacle import DetectorConfig


def get_default_detector_config() -> DetectorConfig:
    return DetectorConfig(
        enable_bilateration=True,
    )



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.surface_velocity._processor import Processor, ProcessorConfig


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client.open(**a121.get_client_args(args))

    sensor_config = a121.SensorConfig(
        profile=a121.Profile.PROFILE_3,
        start_point=180,
        num_points=4,
        step_length=6,
        hwaas=16,
        sweeps_per_frame=128,
        sweep_rate=2000,
        continuous_sweep_mode=True,
        double_buffering=True,
        inter_frame_idle_state=a121.IdleState.READY,
        inter_sweep_idle_state=a121.IdleState.READY,
    )

    metadata = client.setup_session(sensor_config)

    processor_config = ProcessorConfig(
        surface_distance=0.40,
        time_series_length=1024,
    )

    processor = Processor(
        sensor_config=sensor_config,
        metadata=metadata,
        processor_config=processor_config,
    )

    pg_updater = PGUpdater(
        processor_config,
        sensor_config,
    )

    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    client.start_session()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    print(f"Measured distances (m): {np.around(processor.distances, 2)}")

    while not interrupt_handler.got_signal:
        result = client.get_next()
        processor_result = processor.process(result)
        print(
            f"Estimated velocity {np.around(processor_result.estimated_v, 2)} m/s, "
            f"at distance {np.around(processor_result.distance_m, 2)} m"
        )

        try:
            pg_process.put_data(processor_result)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    pg_process.close()
    client.close()


class PGUpdater:
    _VELOCITY_Y_SCALE_MARGIN_M = 0.25

    def __init__(
        self,
        processor_config: ProcessorConfig,
        sensor_config: a121.SensorConfig,
    ):
        self.slow_zone = processor_config.slow_zone
        self.history_length_s = 10
        if sensor_config.frame_rate is None:
            estimated_frame_rate = sensor_config.sweep_rate / sensor_config.sweeps_per_frame
        else:
            estimated_frame_rate = sensor_config.frame_rate

        self.history_length_n = int(np.around(self.history_length_s * estimated_frame_rate))

        self.setup_is_done = False

    def setup(self, win):
        c0_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")

        # Velocity plot

        self.velocity_history_plot = self._create_plot(win, row=0, col=0)
        self.velocity_history_plot.setTitle("Estimated velocity")
        self.velocity_history_plot.setLabel(axis="left", text="Velocity", units="m/s")
        self.velocity_history_plot.setLabel(axis="bottom", text="Time", units="s")
        self.velocity_history_plot.addLegend(labelTextSize="10pt")
        self.velocity_smooth_limits = et.utils.SmoothLimits()

        self.velocity_curve = self.velocity_history_plot.plot(
            pen=et.utils.pg_pen_cycler(0), name="Estimated velocity"
        )

        self.psd_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:13pt;">'
            "{}</span></div>"
        )

        self.distance_text_item = pg.TextItem(
            html=self.psd_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.velocity_history_plot.addItem(self.distance_text_item)

        self.velocity_history = np.zeros(self.history_length_n)

        self.lower_std_history = np.zeros(self.history_length_n)
        self.upper_std_history = np.zeros(self.history_length_n)

        self.lower_std_curve = self.velocity_history_plot.plot()
        self.upper_std_curve = self.velocity_history_plot.plot()

        fbi = pg.FillBetweenItem(
            self.lower_std_curve,
            self.upper_std_curve,
            brush=pg.mkBrush(f"{et.utils.color_cycler(0)}50"),
        )

        self.velocity_history_plot.addItem(fbi)

        # PSD plot

        self.psd_plot = self._create_plot(win, row=1, col=0)
        self.psd_plot.setTitle("PSD<br>(colored area represents the slow zone)")
        self.psd_plot.setLabel(axis="left", text="Power")
        self.psd_plot.setLabel(axis="bottom", text="Velocity", units="m/s")
        self.psd_plot.addLegend(labelTextSize="10pt")

        self.psd_smooth_max = et.utils.SmoothMax(tau_grow=0.5, tau_decay=2.0)
        self.psd_curve = self.psd_plot.plot(pen=et.utils.pg_pen_cycler(0), name="PSD")
        self.psd_threshold = self.psd_plot.plot(pen=c0_dashed_pen, name="Threshold")

        psd_slow_zone_color = et.utils.color_cycler(0)
        psd_slow_zone_color = f"{psd_slow_zone_color}50"
        psd_slow_zone_brush = pg.mkBrush(psd_slow_zone_color)

        self.psd_slow_zone = pg.LinearRegionItem(brush=psd_slow_zone_brush, movable=False)
        self.psd_plot.addItem(self.psd_slow_zone)

        brush = et.utils.pg_brush_cycler(0)
        self.psd_peak_plot_item = pg.PlotDataItem(
            pen=None, symbol="o", symbolSize=8, symbolBrush=brush, symbolPen="k"
        )
        self.psd_plot.addItem(self.psd_peak_plot_item)

        self.psd_plot.setLogMode(x=False, y=True)

    def update(self, processor_result) -> None:
        processor_extra_result = processor_result.extra_result

        lim = self.velocity_smooth_limits.update(processor_result.estimated_v)

        self.velocity_history_plot.setYRange(
            lim[0] - self._VELOCITY_Y_SCALE_MARGIN_M, lim[1] + self._VELOCITY_Y_SCALE_MARGIN_M
        )
        self.velocity_history_plot.setXRange(-self.history_length_s, 0)

        xs = np.linspace(-self.history_length_s, 0, self.history_length_n)

        self.velocity_history = np.roll(self.velocity_history, -1)
        self.velocity_history[-1] = processor_result.estimated_v
        self.velocity_curve.setData(xs, self.velocity_history)

        velocity_html = self.psd_html.format(
            f"Distance {np.around(processor_result.distance_m, 2)} m"
        )
        self.distance_text_item.setHtml(velocity_html)
        self.distance_text_item.setPos(
            -self.history_length_s / 2, lim[1] + self._VELOCITY_Y_SCALE_MARGIN_M
        )

        self.lower_std_history = np.roll(self.lower_std_history, -1)
        self.lower_std_history[-1] = (
            processor_result.estimated_v + 0.5 * processor_extra_result.peak_width
        )
        self.lower_std_curve.setData(xs, self.lower_std_history)

        self.upper_std_history = np.roll(self.upper_std_history, -1)
        self.upper_std_history[-1] = (
            processor_result.estimated_v - 0.5 * processor_extra_result.peak_width
        )
        self.upper_std_curve.setData(xs, self.upper_std_history)

        lim = self.psd_smooth_max.update(processor_extra_result.psd)
        self.psd_plot.setYRange(np.log(0.5), np.log(lim))
        self.psd_plot.setXRange(
            processor_extra_result.max_bin_vertical_vs[0],
            processor_extra_result.max_bin_vertical_vs[-1],
        )
        self.psd_curve.setData(
            processor_extra_result.vertical_velocities, processor_extra_result.psd
        )
        self.psd_threshold.setData(
            processor_extra_result.vertical_velocities, processor_extra_result.psd_threshold
        )
        if processor_extra_result.peak_idx is not None:
            self.psd_peak_plot_item.setData(
                [processor_extra_result.vertical_velocities[processor_extra_result.peak_idx]],
                [processor_extra_result.psd[processor_extra_result.peak_idx]],
            )
        else:
            self.psd_peak_plot_item.clear()

        middle_idx = int(np.around(processor_extra_result.vertical_velocities.shape[0] / 2))
        self.psd_slow_zone.setRegion(
            [
                processor_extra_result.vertical_velocities[middle_idx - self.slow_zone],
                processor_extra_result.vertical_velocities[middle_idx + self.slow_zone],
            ]
        )

    @staticmethod
    def _create_plot(parent: pg.GraphicsLayout, row: int, col: int) -> pg.PlotItem:
        velocity_history_plot = parent.addPlot(row=row, col=col)
        velocity_history_plot.setMenuEnabled(False)
        velocity_history_plot.setMouseEnabled(x=False, y=False)
        velocity_history_plot.hideButtons()
        velocity_history_plot.showGrid(x=True, y=True, alpha=0.5)

        return velocity_history_plot


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2023
# All rights reserved

"""
This example will demonstrate how one can post process already recorded SparseIQ data
with distance or presence (the other algorithms are similar).

Note that some algorithms requires the data to have been recorded with specific settings.
This means that all recorded data cannot be processed by all algorithms.

Usage:

    python path/to/post_process_sparse_iq.py <path/to/h5 file> --processing distance
"""
import argparse
import pathlib

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import distance, presence


def presence_processing(record: a121.H5Record) -> None:
    # User-defined post-processing parameters.
    # See these Presence resources for more information:
    #
    # Presence processor example:
    #   https://github.com/acconeer/acconeer-python-exploration/blob/master/examples/a121/algo/presence/processor.py
    # Presence doc page:
    #   https://docs.acconeer.com/en/latest/exploration_tool/algo/a121/detectors/presence_detection.html
    processor_config = presence.ProcessorConfig(intra_enable=True)

    # Get needed metadata of recorded data from the H5Record
    # If your record contains "extended" data, an error will be raised here.
    record_sensor_config = record.session_config.sensor_config
    metadata = record.metadata

    # Each algorithm has requirements on the data it processes. validate checks these requirements
    processor_config.validate(record_sensor_config)

    # Create a processor with metadata from recording and the user-defined processing parameters
    processor = presence.Processor(
        sensor_config=record_sensor_config,
        metadata=metadata,
        processor_config=processor_config,
    )

    # Post-process all recorded results and print the intra presence score
    for idx, result in enumerate(record.results):
        processor_result = processor.process(result)
        print(f"{idx: >5}: {processor_result.intra_presence_score}")


def distance_processing(record: a121.H5Record) -> None:
    # User-defined post-processing parameters.
    # See these Distance resources for more information:
    #
    # Distance processor example:
    #   https://github.com/acconeer/acconeer-python-exploration/blob/master/examples/a121/algo/distance/processor.py
    # Distance doc page:
    #   https://docs.acconeer.com/en/latest/exploration_tool/algo/a121/detectors/distance_detection.html
    processor_config = distance.ProcessorConfig(
        processor_mode=distance.ProcessorMode.DISTANCE_ESTIMATION,
        threshold_method=distance.ThresholdMethod.CFAR,
        measurement_type=distance.MeasurementType.FAR_RANGE,
        reflector_shape=distance.ReflectorShape.GENERIC,
        threshold_sensitivity=0.5,
        fixed_threshold_value=100.0,
        fixed_strength_threshold_value=0.0,
    )

    # Get needed metadata of recorded data from the H5Record.
    # If your record contains "extended" data, an error will be raised here.
    record_sensor_config = record.session_config.sensor_config
    metadata = record.metadata

    # Each algorithm has requirements on the data it processes. validate checks these requirements
    processor_config.validate(record_sensor_config)

    # Create a processor with metadata from recording and the user-defined processing parameters
    all_subsweep_indexes = list(range(record_sensor_config.num_subsweeps))
    processor = distance.Processor(
        sensor_config=record_sensor_config,
        metadata=metadata,
        processor_config=processor_config,
        subsweep_indexes=all_subsweep_indexes,
    )

    # Post-process all recorded results and print the estimated distances
    for idx, result in enumerate(record.results):
        processor_result = processor.process(result)
        print(f"{idx: >5}: {processor_result.estimated_distances}")


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("file", type=pathlib.Path, help="Path to the recorded .h5 file")
    parser.add_argument(
        "--processing",
        type=str,
        choices=["distance", "presence"],
        help="The post-processing to use",
    )

    args = parser.parse_args()

    with a121.open_record(args.file) as record:
        if args.processing == "distance":
            distance_processing(record)
        elif args.processing == "presence":
            presence_processing(record)
        else:
            print(f"No post-processing available for algorithm {args.processing!r}")


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Optional

import attrs
import h5py
import numpy as np

from PySide6.QtGui import QTransform
from PySide6.QtWidgets import QLabel, QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121, opser
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.a121.algo.obstacle import (
    BilateratorResult,
    DetailedStatus,
    Detector,
    DetectorConfig,
    DetectorContext,
    DetectorResult,
)
from acconeer.exptool.a121.algo.obstacle._detector import _load_algo_data
from acconeer.exptool.a121.algo.obstacle._detector_plugin import (
    ViewPlugin as ObstacleViewPlugin,
)
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    GeneralMessage,
    GroupBox,
    HandledException,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    PluginState,
    PluginStateMessage,
    TwoSensorIdsEditor,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)

from ._configs import get_default_detector_config


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_ids: list[int] = attrs.field(factory=lambda: [2, 3])
    context: DetectorContext = attrs.field(factory=DetectorContext)
    config: DetectorConfig = attrs.field(factory=DetectorConfig)


class PluginPresetId(Enum):
    DEFAULT = auto()


@attrs.mutable(kw_only=True)
class BilaterationPreset:
    detector_config: DetectorConfig = attrs.field()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    bilateration_config: DetectorConfig
    sensor_ids: list[int] = attrs.field(factory=list)
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    def __init__(
        self,
        callback: Callable[[Message], None],
        generation: PluginGeneration,
        key: str,
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._detector_instance: Optional[Detector] = None

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = DetectorConfig.from_json(file["config"][()])
        self.shared_state.context = opser.deserialize(file["context"], DetectorContext)

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState(config=get_default_detector_config())
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            for i in range(len(self.shared_state.sensor_ids)):
                if len(sensor_ids) > 0 and self.shared_state.sensor_ids[i] not in sensor_ids:
                    self.shared_state.sensor_ids[i] = sensor_ids[0]

            self.broadcast()

    @is_task
    def update_config(self, *, config: DetectorConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    @is_task
    def update_processor_config(self, *, config: DetectorConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    @is_task
    def set_preset(self, preset_id: int) -> None:
        if preset_id == PluginPresetId.DEFAULT.value:
            self.restore_defaults()
            self.broadcast()
        else:
            assert False  # <- bomb that will blow up if/when we add a second preset

    @is_task
    def update_sensor_ids(self, *, sensor_ids: list[int]) -> None:
        self.shared_state.sensor_ids = sensor_ids
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())
        opser.serialize(self.shared_state.context, file.create_group("context"))

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config, context = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.context = context

        sensor_ids: list[int] = []
        for entry in record.session_config.groups:
            sensor_ids.extend(entry.keys())
        self.shared_state.sensor_ids = sensor_ids

        self.broadcast()

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client

        self._detector_instance = Detector(
            client=self.client,
            sensor_ids=self.shared_state.sensor_ids,
            detector_config=self.shared_state.config,
            context=self.shared_state.context,
        )
        if recorder is None:
            self._detector_instance.start()
        else:
            obstacle_bilateration_algo_group = recorder.require_algo_group("obstacle_bilateration")
            self._detector_instance.start(recorder, obstacle_bilateration_algo_group)

        self.callback(
            SetupMessage(
                bilateration_config=self.shared_state.config,
                sensor_ids=self.shared_state.sensor_ids,
            )
        )

    def end_session(self) -> None:
        if self._detector_instance is None:
            raise RuntimeError
        if self._recorder is not None:
            self._recorder.close()
        self._detector_instance.stop()

    def get_next(self) -> None:
        if self._detector_instance is None:
            raise RuntimeError
        detector_result = self._detector_instance.get_next()

        assert self.client is not None
        self.callback(backend.PlotMessage(result=detector_result))

    @is_task
    def calibrate_detector(self) -> None:
        if self._started:
            raise RuntimeError

        if self.client is None:
            raise RuntimeError

        if not self.client.connected:
            raise RuntimeError

        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

        try:
            self._detector_instance = Detector(
                client=self.client,
                sensor_ids=self.shared_state.sensor_ids,
                detector_config=self.shared_state.config,
            )
            self._detector_instance.calibrate_detector()
        except Exception as exc:
            raise HandledException("Failed to calibrate detector") from exc
        finally:
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

        self.shared_state.context = self._detector_instance.context

        self.broadcast()


PLOT_HISTORY_FRAMES = 50


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(
                message.bilateration_config,
                message.sensor_ids,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(detector_result=self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        detector_config: DetectorConfig,
        sensor_ids: list[int],
    ) -> None:
        self.plot_layout.clear()

        self.detector_config = detector_config
        self.sensor_ids = sensor_ids

        self.fftmap_plots: list[pg.PlotItem] = []
        self.fftmap_images: list[pg.ImageItem] = []
        self.range_hist_plots: list[pg.PlotItem] = []
        self.range_hist_curves: list[pg.PlotDataItem] = []
        self.angle_hist_curves: list[pg.PlotDataItem] = []
        self.close_proximity_text_items: list[pg.TextItem] = []

        self.obst_vel_ys = 2 * [np.nan * np.ones(PLOT_HISTORY_FRAMES)]
        self.obst_dist_ys = 2 * [np.nan * np.ones(PLOT_HISTORY_FRAMES)]
        self.obst_bil_ys = np.nan * np.ones(PLOT_HISTORY_FRAMES)
        self.hist_x = np.linspace(-100, 0, PLOT_HISTORY_FRAMES)  # TODO: should no be hard coded

        win = self.plot_layout

        for i_s in range(2):
            p = win.addPlot(row=0, col=i_s, title=f"FFT Map, sensor {self.sensor_ids[i_s]} ")

            im = pg.ImageItem(autoDownsample=True)
            im.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
            self.fftmap_images.append(im)

            p.setLabel("bottom", "Distance (cm)")
            p.setLabel("left", "Velocity (cm/s)")
            p.addItem(im)

            self.fftmap_plots.append(p)

            self.angle_hist = pg.PlotItem(title="Angle/velocity history")
            self.angle_hist.showGrid(x=True, y=True)
            self.angle_hist.setLabel("bottom", "Time (frames)")
            self.angle_hist.setLabel("left", "velocity (cm/s)")
            self.angle_hist.setXRange(-100, 0)
            self.angle_hist.addLegend()
            self.angle_hist_curves.append(self.angle_hist.plot(symbolSize=5, symbol="o"))

            sublayout = win.addLayout(
                row=1,
                col=i_s,
                colspan=1,
            )

            sublayout.layout.setColumnStretchFactor(0, 1)
            sublayout.addItem(self.angle_hist, row=0, col=0)

            range_hist = pg.PlotItem(title="Range history")
            range_hist.showGrid(x=True, y=True)
            range_hist.setLabel("bottom", "Time (frames)")
            range_hist.setLabel("left", "Range (cm)")
            range_hist.setXRange(-100, 0)
            range_hist.addLegend()
            self.range_hist_plots.append(range_hist)
            self.range_hist_curves.append(range_hist.plot(symbolSize=5, symbol="o"))

            close_proximity_html = (
                '<p style="text-align: center; color: white; font-size: 10pt;">{}</p>'.format(
                    "Close proximity detection",
                )
            )

            close_proximity_text_item = pg.TextItem(
                html=close_proximity_html,
                fill=pg.mkColor(0xFF, 0x7F, 0x0E),
                anchor=(0.5, 0),
            )
            self.close_proximity_text_items.append(close_proximity_text_item)
            range_hist.addItem(close_proximity_text_item)
            close_proximity_text_item.hide()

            sublayout = win.addLayout(
                row=2,
                col=i_s,
                colspan=1,
            )

            sublayout.layout.setColumnStretchFactor(0, 1)
            sublayout.addItem(range_hist, row=0, col=0)

        self.bil_hist_plot = pg.PlotItem(title="Bilateration history")

        self.bil_hist_plot.showGrid(x=True, y=True)
        self.bil_hist_plot.setLabel("bottom", "Time (frames)")
        self.bil_hist_plot.setLabel("left", "Bilateration angle (deg)")
        self.bil_hist_plot.setXRange(-100, 0)
        self.bil_hist_plot.setYRange(-90, 90)
        self.bil_hist_plot.addLegend()

        self.bil_hist_curve = self.bil_hist_plot.plot(pen=et.utils.pg_pen_cycler(1))

        sublayout = win.addLayout(row=3, col=0, colspan=2)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.bil_hist_plot, row=0, col=0)

        self.setup_is_done = True

    def draw_plot_job(self, *, detector_result: DetectorResult) -> None:
        # Plot sweep data from both distance detectors.

        for i_s in range(2):
            pr = detector_result.processor_results[self.sensor_ids[i_s]]
            er = pr.subsweeps_extra_results[0]

            fftmap = pr.subsweeps_extra_results[0].fft_map

            spf = fftmap.shape[0]
            r = 100 * pr.subsweeps_extra_results[0].r

            transform = QTransform()
            transform.translate(
                r[0],
                -100 * pr.extra_result.dv * spf / 2 - 0.5 * 100 * pr.extra_result.dv,
            )
            transform.scale(r[1] - r[0], 100 * pr.extra_result.dv)

            self.fftmap_images[i_s].setTransform(transform)

            self.fftmap_images[i_s].updateImage(
                np.fft.fftshift(fftmap, 0).T,
                levels=(0, 1.05 * np.max(fftmap)),
            )

            v = pr.targets[0].velocity if pr.targets else np.nan

            self.obst_vel_ys[i_s] = np.roll(self.obst_vel_ys[i_s], -1)
            self.obst_vel_ys[i_s][-1] = 100 * v  # m/s -> cm/s

            if np.isnan(self.obst_vel_ys[i_s]).all():
                self.angle_hist_curves[i_s].setVisible(False)
            else:
                self.angle_hist_curves[i_s].setVisible(True)
                self.angle_hist_curves[i_s].setData(
                    self.hist_x, self.obst_vel_ys[i_s], connect="finite"
                )

            r_targets = pr.targets[0].distance if pr.targets else np.nan

            self.obst_dist_ys[i_s] = np.roll(self.obst_dist_ys[i_s], -1)
            self.obst_dist_ys[i_s][-1] = 100 * r_targets  # m -> cm

            min_range = 100 * er.r[0]
            max_range = 100 * er.r[-1]
            self.range_hist_plots[i_s].setYRange(min_range, max_range)

            if np.isnan(self.obst_dist_ys[i_s]).all():
                self.range_hist_curves[i_s].setVisible(False)
            else:
                self.range_hist_curves[i_s].setVisible(True)
                self.range_hist_curves[i_s].setData(
                    self.hist_x, self.obst_dist_ys[i_s], connect="finite"
                )

            if detector_result.close_proximity_trig:
                close_prox = detector_result.close_proximity_trig[self.sensor_ids[i_s]]
                self.close_proximity_text_items[i_s].setPos(-50, max_range)
                self.close_proximity_text_items[i_s].setVisible(close_prox)

        assert isinstance(detector_result.bilateration_result, BilateratorResult)

        beta = (
            detector_result.bilateration_result.beta_degs[0]
            if detector_result.bilateration_result.beta_degs
            else np.nan
        )

        self.obst_bil_ys = np.roll(self.obst_bil_ys, -1)
        self.obst_bil_ys[-1] = beta

        if np.isnan(self.obst_bil_ys).all():
            self.bil_hist_curve.setVisible(False)
        else:
            self.bil_hist_curve.setVisible(True)
            self.bil_hist_curve.setData(self.hist_x, self.obst_bil_ys, connect="finite")


class ViewPlugin(A121ViewPluginBase):
    TEXT_MSG_MAP = {
        DetailedStatus.OK: "Ready to start.",
        DetailedStatus.END_LESSER_THAN_START: "'Range end' point must be greater than 'Range "
        + "start'.",
        DetailedStatus.SENSOR_IDS_NOT_UNIQUE: "Select two different sensor IDs.",
        DetailedStatus.CONTEXT_MISSING: "Run detector calibration.",
        DetailedStatus.CALIBRATION_MISSING: "Run detector calibration.",
        DetailedStatus.CONFIG_MISMATCH: (
            "Current configuration does not match the configuration "
            + "used during detector calibration. Run detector calibration."
        ),
    }

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        self.calibrate_detector_button = QPushButton(icons.CALIBRATE(), "Calibrate detector")
        self.calibrate_detector_button.clicked.connect(self._on_calibrate_detector)

        self.defaults_button = QPushButton(icons.RESTORE(), "Reset settings and calibrations")
        self.defaults_button.clicked.connect(self._send_defaults_request)

        self.message_box = QLabel(self.sticky_widget)
        self.message_box.setWordWrap(True)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        button_group.layout().addWidget(self.calibrate_detector_button, 1, 0, 1, -1)
        button_group.layout().addWidget(self.defaults_button, 3, 0, 1, -1)
        button_group.layout().addWidget(self.message_box, 4, 0, 1, -1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.grid("Sensor selection", parent=self.scrolly_widget)
        self.two_sensor_id_editor = TwoSensorIdsEditor(name_label_texts=["Left", "Right"])
        sensor_selection_group.layout().addWidget(self.two_sensor_id_editor, 0, 0)
        scrolly_layout.addWidget(sensor_selection_group)

        self.config_editor = AttrsConfigEditor(
            title="Obstacle detector parameters",
            config_type=DetectorConfig,
            factory_mapping=self._get_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.two_sensor_id_editor.sig_update.connect(self._on_sensor_ids_update)
        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_processor_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {}

    @classmethod
    def _get_pidget_mapping(cls) -> PidgetFactoryMapping:
        obstacle_dict = dict(ObstacleViewPlugin.get_pidget_mapping())
        bilateration_dict = {
            "bilateration_sensor_spacing_m": pidgets.FloatPidgetFactory(
                name_label_text="Sensor spacing:",
                suffix=" m",
                decimals=3,
            ),
        }
        obstacle_dict.update(bilateration_dict)
        return obstacle_dict

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
            self.message_box.setText("")
        else:
            self.config_editor.set_data(state.config)
            self.two_sensor_id_editor.set_data(state.sensor_ids)

            detector_status = Detector.get_detector_status(
                state.config, state.context, state.sensor_ids
            )

            self.message_box.setText(self.TEXT_MSG_MAP[detector_status.detector_state])

            validation_results = state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(validation_results)
            not_handled = self.misc_error_view.handle_validation_results(not_handled)
            assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.two_sensor_id_editor.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[
                self.config_editor,
                self.defaults_button,
                self.two_sensor_id_editor,
            ],
        )

        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

        state = app_model.backend_plugin_state

        if state is None:
            detector_ready = False
            state_valid = False
        else:
            detector_status = Detector.get_detector_status(
                state.config, state.context, state.sensor_ids
            )
            detector_ready = detector_status.ready_to_start

            state_valid = (
                self._config_valid(state)
                and self.config_editor.is_ready
                and detector_status.detector_state != DetailedStatus.SENSOR_IDS_NOT_UNIQUE
            )

        self.calibrate_detector_button.setEnabled(
            visual_policies.start_button_enabled(app_model, extra_condition=state_valid)
        )
        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=detector_ready and state_valid
            )
        )

    def _config_valid(self, state: SharedState) -> bool:
        try:
            state.config.validate()
        except a121.ValidationResult:
            return False
        else:
            return True

    def _on_config_update(self, config: DetectorConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_processor_config_update(self, config: DetectorConfig) -> None:
        BackendPlugin.update_processor_config.rpc(self.app_model.put_task, config=config)

    def _on_sensor_ids_update(self, sensor_ids: list[int]) -> None:
        BackendPlugin.update_sensor_ids.rpc(self.app_model.put_task, sensor_ids=sensor_ids)

    def _on_calibrate_detector(self) -> None:
        BackendPlugin.calibrate_detector.rpc(self.app_model.put_task)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


OBSTACLE_BILATERATION_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="obstacle_bilateration",
    title="Obstacle Bilateration",
    docs_link="https://docs.acconeer.com/en/latest/exploration_tool/algo/a121/examples/obstacle_bilateration.html",
    description="Use two sensors to estimate distance and two angles.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.distance import (
    PeakSortingMethod,
    ReflectorShape,
    ThresholdMethod,
)
from acconeer.exptool.a121.algo.distance._processors import (
    DEFAULT_FIXED_AMPLITUDE_THRESHOLD_VALUE,
    DEFAULT_FIXED_STRENGTH_THRESHOLD_VALUE,
)
from acconeer.exptool.a121.algo.tank_level import RefApp
from acconeer.exptool.a121.algo.tank_level._ref_app import RefAppConfig


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    # Setup the configurations
    # Detailed at https://docs.acconeer.com/en/latest/exploration_tool/algo/a121/ref_apps/tank_level.html

    # Sensor selections
    sensor = 1

    # Tank level configurations
    ref_app_config = RefAppConfig(
        start_m=0.03,
        end_m=0.5,
        max_step_length=2,
        max_profile=a121.Profile.PROFILE_2,
        close_range_leakage_cancellation=True,
        signal_quality=20,
        update_rate=None,
        median_filter_length=5,
        num_medians_to_average=5,
        threshold_method=ThresholdMethod.CFAR,
        reflector_shape=ReflectorShape.PLANAR,
        peaksorting_method=PeakSortingMethod.CLOSEST,
        num_frames_in_recorded_threshold=50,
        fixed_threshold_value=DEFAULT_FIXED_AMPLITUDE_THRESHOLD_VALUE,  # float
        fixed_strength_threshold_value=DEFAULT_FIXED_STRENGTH_THRESHOLD_VALUE,  # float
        threshold_sensitivity=0.0,  # float
    )

    # End setup configurations

    # Preparation for client
    client = a121.Client.open(**a121.get_client_args(args))

    # Preparation for reference application processor
    ref_app = RefApp(client=client, sensor_id=sensor, config=ref_app_config)
    ref_app.calibrate()
    ref_app.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        processed_data = ref_app.get_next()
        try:
            if processed_data.level is not None:
                print("Tank level result " + str(processed_data.level))
        except et.PGProccessDiedException:
            break

    ref_app.stop()
    client.close()
    print("Disconnecting...")


if __name__ == "__main__":
    main()
